###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""The ErrorLogging module is used to perform error analysis using AppConfig
utilities. This occurs at the end of each workflow step such that the
step_commons dictionary can be utilized.

Since not all projects are instrumented to work with the AppConfig error
suite any failures will not be propagated to the workflow.
"""
import os

from DIRAC import S_OK, S_ERROR, gLogger

import LHCbDIRAC.Core.Utilities.LogErr as LogErr
from LHCbDIRAC.Workflow.Modules.ModuleBase import ModuleBase


class ErrorLogging(ModuleBase):
    #############################################################################

    def __init__(self):
        """c'tor."""

        self.log = gLogger.getSubLogger("ErrorLogging")
        super().__init__(self.log)

        # Internal parameters
        self.errorLogNameHTML = ""
        self.errorLogNamejson = ""

    #############################################################################

    def _resolveInputVariables(self):
        """By convention the module input parameters are resolved here."""
        super()._resolveInputVariables()
        super()._resolveInputStep()

        self.errorLogNamejson = "%d_Errors_%s.json" % (self.jobID, self.applicationName)

    #############################################################################

    def execute(
        self,
        production_id=None,
        prod_job_id=None,
        wms_job_id=None,
        workflowStatus=None,
        stepStatus=None,
        wf_commons=None,
        step_commons=None,
        step_number=None,
        step_id=None,
    ):
        """Main execution function. Always return S_OK() because we don't want the
        job execution result to depend on retrieving errors from logs.

        This module will run regardless of the workflow status.
        """

        try:
            super().execute(
                production_id,
                prod_job_id,
                wms_job_id,
                workflowStatus,
                stepStatus,
                wf_commons,
                step_commons,
                step_number,
                step_id,
            )

            self._resolveInputVariables()

            if self.applicationName.lower() not in ("gauss", "boole"):
                self.log.info("Not Gauss nor Boole, exiting")
                return S_OK()

            self.log.info(
                "Executing ErrorLogging module for: %s %s %s"
                % (self.applicationName, self.applicationVersion, self.applicationLog)
            )
            if not os.path.exists(self.applicationLog):
                self.log.info(f"Application log file from previous module not found locally: {self.applicationLog}")
                return S_OK()

            # Now really running
            # self.step_commons['extraPackages'] is something like 'AppConfig.v3r360;TurboStreamProd.v4r2p9;ProdConf'
            result = LogErr.readLogFile(
                logFile=self.applicationLog,
                jobID=self.prod_job_id,
                prodID=self.production_id,
                wmsID=self.jobID,
                application=self.applicationName,
                applicationVersion=self.applicationVersion,
                name=self.errorLogNamejson,
            )

            if not result["OK"]:
                self.log.error(
                    "Error logging for %s %s step %s completed with errors:"
                    % (self.applicationName, self.applicationVersion, self.step_number)
                )
                self.log.warn("Exiting without affecting workflow status")
                return S_OK()

            if not os.path.exists(self.errorLogNamejson):
                self.log.error(f"{self.errorLogNamejson} not found locally, exiting without affecting workflow status")
                return S_OK()

            self.log.info(
                "Error logging for %s %s step %s completed successfully:"
                % (self.applicationName, self.applicationVersion, self.step_number)
            )

            return S_OK()

        except Exception as e:
            self.log.exception("Failure in ErrorLogging execute module", lException=e)
            return S_ERROR("Error in ErrorLogging module")

        finally:
            super().finalize()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Unit test of ConsistencyChecks."""
# pylint: disable=invalid-name,missing-docstring,protected-access
from unittest.mock import Mock

import pytest

from LHCbDIRAC.BookkeepingSystem.Client.test.mock_BookkeepingClient import bkc_mock
from LHCbDIRAC.DataManagementSystem.Client import ConsistencyChecks as ConsistencyChecksModule

FILETYPES = [
    ["SEMILEPTONIC.DST", "LOG", "RAW"],
    ["SEMILEPTONIC.DST", "LOG", "RAW"],
    ["SEMILEPTONIC.DST"],
    ["SEMILEPTONIC.DST"],
]


@pytest.fixture
def cc_mock(monkeypatch):
    dmMock = Mock()
    dmMock.getReplicas.return_value = {
        "OK": True,
        "Value": {"Successful": {"bb.raw": "metadataPippo"}, "Failed": {}},
    }

    cc = ConsistencyChecksModule.ConsistencyChecks(transClient=Mock(), dm=dmMock, bkClient=bkc_mock)

    cc.fileTypesExcluded = ["LOG"]
    cc.prod = 0
    monkeypatch.setattr(ConsistencyChecksModule, "BookkeepingClient", lambda: bkc_mock)
    yield cc


def test_selectByFileType(cc_mock):
    lfnDicts = [
        {
            "aa.raw": {
                "bb.raw": {"FileType": "RAW", "RunNumber": 97019},
                "bb.log": {"FileType": "LOG"},
                "/bb/pippo/aa.dst": {"FileType": "DST"},
                "/lhcb/1_2_1.Semileptonic.dst": {"FileType": "SEMILEPTONIC.DST"},
            },
            "cc.raw": {
                "dd.raw": {"FileType": "RAW", "RunNumber": 97019},
                "bb.log": {"FileType": "LOG"},
                "/bb/pippo/aa.dst": {"FileType": "LOG"},
                "/lhcb/1_1.semileptonic.dst": {"FileType": "SEMILEPTONIC.DST"},
            },
        },
        {"aa.raw": {"/bb/pippo/aa.dst": {"FileType": "LOG"}, "bb.log": {"FileType": "LOG"}}},
        {
            "aa.raw": {
                "bb.raw": {"FileType": "RAW", "RunNumber": 97019},
                "bb.log": {"FileType": "LOG"},
                "/bb/pippo/aa.dst": {"FileType": "DST"},
                "/lhcb/1_2_1.Semileptonic.dst": {"FileType": "SEMILEPTONIC.DST"},
            },
            "cc.raw": {
                "dd.raw": {"FileType": "RAW", "RunNumber": 97019},
                "bb.log": {"FileType": "LOG"},
                "/bb/pippo/aa.dst": {"FileType": "LOG"},
                "/lhcb/1_1.semileptonic.dst": {"FileType": "SEMILEPTONIC.DST"},
            },
        },
        {"aa.raw": {"/bb/pippo/aa.dst": {"FileType": "LOG"}, "bb.log": {"FileType": "LOG"}}},
    ]

    lfnDictsExpected = [
        {
            "aa.raw": {
                "/lhcb/1_2_1.Semileptonic.dst": {"FileType": "SEMILEPTONIC.DST"},
                "bb.log": {"FileType": "LOG"},
                "bb.raw": {"RunNumber": 97019, "FileType": "RAW"},
            },
            "cc.raw": {
                "dd.raw": {"RunNumber": 97019, "FileType": "RAW"},
                "bb.log": {"FileType": "LOG"},
                "/bb/pippo/aa.dst": {"FileType": "LOG"},
                "/lhcb/1_1.semileptonic.dst": {"FileType": "SEMILEPTONIC.DST"},
            },
        },
        {"aa.raw": {"/bb/pippo/aa.dst": {"FileType": "LOG"}, "bb.log": {"FileType": "LOG"}}},
        {
            "aa.raw": {"/lhcb/1_2_1.Semileptonic.dst": {"FileType": "SEMILEPTONIC.DST"}},
            "cc.raw": {"/lhcb/1_1.semileptonic.dst": {"FileType": "SEMILEPTONIC.DST"}},
        },
        {},
    ]

    # testing various cases
    for fileType, lfnDict, lfnDictExpected in zip(FILETYPES, lfnDicts, lfnDictsExpected):
        cc_mock.fileType = fileType

        res = cc_mock._selectByFileType(lfnDict)

        assert res == lfnDictExpected

        res = cc_mock._selectByFileType(lfnDict)
        assert res == lfnDictExpected


def test_getFileTypesCount(cc_mock):
    lfnDict = {"aa.raw": {"bb.log": {"FileType": "LOG"}, "/bb/pippo/aa.dst": {"FileType": "DST"}}}
    res = cc_mock._getFileTypesCount(lfnDict)
    resExpected = {"aa.raw": {"DST": 1, "LOG": 1}}
    assert res == resExpected

    lfnDict = {
        "aa.raw": {
            "bb.log": {"FileType": "LOG"},
            "/bb/pippo/aa.dst": {"FileType": "DST"},
            "/bb/pippo/cc.dst": {"FileType": "DST"},
        }
    }
    res = cc_mock._getFileTypesCount(lfnDict)
    resExpected = {"aa.raw": {"DST": 2, "LOG": 1}}
    assert res == resExpected


def test_getDescendants(cc_mock):
    cc_mock.fileType = ["SEMILEPTONIC.DST", "LOG", "RAW"]
    res = cc_mock.getDescendants(["aa.raw"])
    (
        filesWithDescendants,
        filesWithoutDescendants,
        filesWitMultipleDescendants,
        descendants,
        inFCNotInBK,
        inBKNotInFC,
        removedFiles,
        inFailover,
    ) = res
    assert {k: set(v) for k, v in filesWithDescendants.items()} == {"aa.raw": {"bb.log", "bb.raw"}}
    assert filesWithoutDescendants == {}
    assert filesWitMultipleDescendants == {}
    assert descendants == ["bb.raw"]
    assert inFCNotInBK == []
    assert inBKNotInFC == []
    assert removedFiles == []
    assert inFailover == []



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""stores the input files."""
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Job.File import File


class InputFile(File):
    """InputFile class."""

    def writeToXML(self):
        """creates an xml string."""
        return '  <InputFile    Name="' + self.name + '"/>\n'



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Production API.

A production is an augmented version of an LHCbJob

Notes:

- Supports all workflows
- create() method that takes a workflow or Production object
  and publishes to the production management system, in addition this
  can automatically construct and publish the BK pass info and transformations
- Uses __getOutputLFNs() function to add production output directory parameter
"""
import json
import shutil
import re
import os
import pprint

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Workflow.Workflow import Workflow, fromXMLString
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getSites, getSiteTier
from DIRAC.Workflow.Utilities.Utils import getStepDefinition

from LHCbDIRAC.Core.Utilities.ProductionData import preSubmissionLFNs
from LHCbDIRAC.Interfaces.API.LHCbJob import LHCbJob
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequestClient import ProductionRequestClient
from LHCbDIRAC.TransformationSystem.Client.Transformation import Transformation


class Production:
    """Production uses an LHCbJob object, as well as few clients."""

    #############################################################################

    def __init__(self, script=None):
        """Instantiates the Workflow object and some default parameters."""

        self.LHCbJob = LHCbJob(script)
        self.bkkClient = BookkeepingClient()
        self.transformation = Transformation()
        self.opsHelper = Operations()

        self.bkSteps = {}
        self.prodGroup = ""
        self.plugin = ""
        self.inputFileMask = ""
        self.inputBKSelection = {}
        self.jobFileGroupSize = 0
        self.ancestorProduction = 0
        self.transformationFamily = 0
        self.priority = 1
        self.gaudiSteps = []

        # a dictionary with 'FileType':'SEName'. All the file types should be present. HISTs are present by default
        histogramSE = self.opsHelper.getValue("Productions/HistogramSE", "CERN-HIST")
        histoTypes = self.opsHelper.getValue(
            "Productions/HistogramTypes", ["HIST", "BRUNELHIST", "DAVINCIHIST", "GAUSSHIST"]
        )
        self.outputSEs = {ht: histogramSE for ht in histoTypes}

        if not script:
            self.__setDefaults()

    #############################################################################

    def __setDefaults(self):
        """Sets some default parameters."""

        self.LHCbJob.stepCount = 0
        self.LHCbJob.setOutputSandbox(
            self.opsHelper.getValue("Productions/inOutputSandbox", ["std.out", "std.err", "*.log"])
        )
        self.setJobParameters(
            {
                "Type": "MCSimulation",
                "Platform": "ANY",
                "CPUTime": "1000000",
                "LogLevel": "verbose",
                "JobGroup": "@{PRODUCTION_ID}",
            }
        )

        self.setFileMask("")

        # General workflow parameters
        self.setParameter("PRODUCTION_ID", "string", "00012345", "ProductionID")
        self.setParameter("JOB_ID", "string", "00006789", "ProductionJobID")
        self.setParameter("poolXMLCatName", "string", "pool_xml_catalog.xml", "POOLXMLCatalogName")
        self.setParameter("outputMode", "string", "Any", "SEResolutionPolicy")
        self.setParameter("outputDataFileMask", "string", "", "outputDataFileMask")

        # BK related parameters
        self.setParameter("configName", "string", "MC", "ConfigName")
        self.setParameter("configVersion", "string", "2009", "ConfigVersion")
        self.setParameter("conditions", "string", "", "SimOrDataTakingCondsString")

    #############################################################################

    def setJobParameters(self, parametersDict):
        """Set an (LHCb)Job parameter.

        The parametersDict is in the form {'parameterName': 'value'} Each
        parameter calls LHCbJob.setparameterName(value)
        """

        for parameter in parametersDict:
            getattr(self.LHCbJob, "set" + parameter)(parametersDict[parameter])

    #############################################################################

    def setParameter(self, name, parameterType, parameterValue, description):
        """Set parameters checking in CS in case some defaults need to be
        changed."""
        proposedParam = self.opsHelper.getValue(f"Productions/{name}", "")
        if proposedParam:
            gLogger.debug(f"Setting {name} from CS defaults = {proposedParam}")
            self.LHCbJob._addParameter(
                self.LHCbJob.workflow, name, parameterType, proposedParam, description
            )  # pylint: disable=protected-access
        else:
            gLogger.debug(f"Setting parameter {name} = {parameterValue}")
            self.LHCbJob._addParameter(
                self.LHCbJob.workflow, name, parameterType, parameterValue, description
            )  # pylint: disable=protected-access

    #############################################################################

    @staticmethod
    def __checkArguments(extraPackages, optionsFile):
        """Checks for typos in the structure of standard arguments to workflows.

        In case of any non-standard settings will raise an exception
        preventing creation of the production. Must be called after setting
        the first event type of the production.
        """
        if not extraPackages:
            extraPackages = []

        if not optionsFile:
            optionsFile = []

        if extraPackages:
            if not re.search(";", extraPackages):
                extraPackages = [extraPackages]
            else:
                extraPackages = extraPackages.split(";")
        if optionsFile:
            # ????
            if not re.search(";", optionsFile):
                optionsFile = [optionsFile]

        for package in extraPackages:
            gLogger.verbose(f"Checking extra package: {package}")
            if not re.search(".", package):
                raise TypeError(f"Must have extra packages in the following format 'Name.Version' not {package}")

        gLogger.verbose("Extra packages and event type options are correctly specified")
        return S_OK()

    #############################################################################

    def addApplicationStep(self, stepDict, inputData=None, modulesList=None, extraModulesList=None):
        """Adds an application step to the workflow.

        stepDict contains everything that is in the step, for this production, e.g.::

          {'ApplicationName': 'DaVinci', 'Usable': 'Yes', 'StepId': 13718, 'ApplicationVersion': 'v28r3p1',
           'ExtraPackages': 'AppConfig.v3r104', 'StepName': 'Stripping14-Merging', 'ExtraOptions': '',
           'EventTimeout': '1000', 'ProcessingPass': 'Merging', 'Visible': 'N', 'OptionsFormat': '',
           'OptionFiles': '$APPCONFIGOPTS/Merging/DV-Stripping14-Merging.py',
           'DDDB': 'head-20110302', 'CONDDB': 'head-20110407', 'DQTag': '',
           'isMulticore': 'N', 'SystemConfig': '', 'mcTCK': '',
           'fileTypesIn': ['SDST'],
           'visibilityFlag': [{'Visible': 'Y', 'FileType': 'BHADRON.DST'}],
           'fileTypesOut': ['BHADRON.DST', 'CALIBRATION.DST', 'CHARM.MDST', 'CHARMCOMPLETEEVENT.DST']}

        Note: this step treated here does not necessarily corresponds to a step of the BKK:
        the case where they might be different is the merging case.

        :param dict stepDict: contains everything that is in the step, for this production
        :param str inputData: the input data of the step. Either None, or 'previousStep', or the input to the step
        :param list modulesList: the list of module names (str) this step is made of. If None, a default is taken.

        :returns: the name (str) of the step added
        """

        appName = stepDict["ApplicationName"]
        appVersion = stepDict["ApplicationVersion"]
        optionsFile = stepDict["OptionFiles"]
        stepID = stepDict["StepId"]
        stepName = stepDict["StepName"]
        stepVisible = stepDict["Visible"]
        extraPackages = stepDict["ExtraPackages"]
        optionsLine = stepDict["ExtraOptions"]
        eventTimeout = stepDict["EventTimeout"]
        fileTypesIn = stepDict["fileTypesIn"]
        fileTypesOut = stepDict["fileTypesOut"]
        stepPass = stepDict["ProcessingPass"]
        optionsFormat = stepDict["OptionsFormat"]
        dddbOpt = stepDict["DDDB"]
        conddbOpt = stepDict["CONDDB"]
        dqOpt = stepDict["DQTag"]
        multicore = stepDict["isMulticore"]
        sysConfig = stepDict["SystemConfig"]
        outputVisibility = stepDict["visibilityFlag"]
        if sysConfig == "None" or sysConfig == "NULL" or not sysConfig or sysConfig is None:
            sysConfig = "ANY"
        mcTCK = stepDict["mcTCK"]

        if extraPackages:
            if isinstance(extraPackages, list):
                extraPackages = ";".join(extraPackages)
            if not isinstance(extraPackages, str):
                raise TypeError("extraPackages is not a string (nor a list)")
            if "," in extraPackages:
                extraPackages = ";".join(extraPackages.split(","))
            extraPackages = extraPackages.replace(" ", "")

        self.__checkArguments(extraPackages, optionsFile)

        try:
            if not dddbOpt.lower() == "global":
                gLogger.verbose(f"Specific DDDBTag setting found for {appName} step, setting to: {dddbOpt}")
                dddbOpt = dddbOpt.replace(" ", "")
        except AttributeError:
            pass
        try:
            if not conddbOpt.lower() == "global":
                gLogger.verbose(f"Specific CondDBTag setting found for {appName} step, setting to: {conddbOpt}")
                conddbOpt = conddbOpt.replace(" ", "")
        except AttributeError:
            pass
        try:
            if not dqOpt.lower() == "global":
                gLogger.verbose(f"Specific DQTag setting found for {appName} step, setting to: {dqOpt}")
                dqOpt = dqOpt.replace(" ", "")
        except AttributeError:
            pass

        # starting real stuff
        self.LHCbJob.stepCount += 1

        if "Gaudi_App_Step" not in self.LHCbJob.workflow.step_definitions:
            gLogger.debug(f"Determining the modules of the steps (modulesList = {modulesList})")
            if modulesList is None:  # we assume it's a standard list of modules for Gaudi steps
                gaudiPath = "Productions/GaudiStep_Modules"
                modulesList = self.opsHelper.getValue(
                    gaudiPath,
                    ["GaudiApplication", "AnalyseXMLSummary", "ErrorLogging", "BookkeepingReport", "StepAccounting"],
                )
            if extraModulesList:
                modulesList += extraModulesList

            # pName, pType, pValue, pDesc
            parametersList = [
                ["inputData", "string", "", "StepInputData"],
                ["inputDataType", "string", "", "InputDataType"],
                ["applicationName", "string", "", "ApplicationName"],
                ["applicationVersion", "string", "", "ApplicationVersion"],
                ["applicationType", "string", "", "ApplicationType"],
                ["optionsFile", "string", "", "OptionsFile"],
                ["extraOptionsLine", "string", "", "extraOptionsLines"],
                ["eventTimeout", "string", "", "EventTimeout"],
                ["listoutput", "list", [], "StepOutputList"],
                ["extraPackages", "string", "", "ExtraPackages"],
                ["BKStepID", "string", "", "BKKStepID"],
                ["StepProcPass", "string", "", "StepProcessingPass"],
                ["optionsFormat", "string", "", "ProdConf configuration"],
                ["CondDBTag", "string", "", "ConditionDatabaseTag"],
                ["DDDBTag", "string", "", "DetDescTag"],
                ["DQTag", "string", "", "DataQualityTag"],
                ["multiCore", "string", "", "MultiCore Flag"],
                ["SystemConfig", "string", "", "system config"],
                ["mcTCK", "string", "", "TCK to be simulated"],
            ]

            gaudiStepDef = getStepDefinition(
                "Gaudi_App_Step",
                modulesNameList=modulesList,
                importLine="LHCbDIRAC.Workflow.Modules",
                parametersList=parametersList,
            )
            self.LHCbJob.workflow.addStep(gaudiStepDef)

        # create the step instance add it to the wf, and return it
        name = f"{appName}_{self.LHCbJob.stepCount}"
        gaudiStepInstance = self.LHCbJob.workflow.createStepInstance("Gaudi_App_Step", name)

        valuesToSet = [
            ["applicationName", appName],
            ["applicationVersion", appVersion],
            ["optionsFile", optionsFile],
            ["extraOptionsLine", optionsLine],
            ["eventTimeout", eventTimeout],
            ["extraPackages", extraPackages],
            ["BKStepID", str(stepID)],
            ["StepProcPass", stepPass],
            ["optionsFormat", optionsFormat],
            ["CondDBTag", conddbOpt],
            ["DDDBTag", dddbOpt],
            ["DQTag", dqOpt],
            ["multiCore", multicore],
            ["SystemConfig", sysConfig],
            ["mcTCK", mcTCK],
        ]

        if fileTypesIn:
            valuesToSet.append(["inputDataType", ";".join(ftIn.upper() for ftIn in fileTypesIn)])

        if inputData is None:
            gLogger.verbose(f"{appName} step has no data requirement or is linked to the overall input data")
            gaudiStepInstance.setLink("inputData", "self", "InputData")
        elif inputData == "previousStep":
            gLogger.verbose("Taking input data as output from previous Gaudi step")
            valuesToSet.append(["inputData", inputData])
        else:
            gLogger.verbose("Assume input data requirement should be added to job")
            self.LHCbJob.setInputData(inputData)
            gaudiStepInstance.setLink("inputData", "self", "InputData")

        for pName, value in valuesToSet:
            if value:
                gaudiStepInstance.setValue(pName, value)

        outputFilesList = self._constructOutputFilesList(fileTypesOut)
        gaudiStepInstance.setValue("listoutput", (outputFilesList))

        # to construct the BK processing pass structure, starts from step '0'
        stepIDInternal = f"Step{self.LHCbJob.stepCount - 1}"
        bkOptionsFile = optionsFile

        stepBKInfo = {
            "ApplicationName": appName,
            "ApplicationVersion": appVersion,
            "OptionFiles": bkOptionsFile,
            "DDDb": dddbOpt,
            "CondDb": conddbOpt,
            "DQTag": dqOpt,
            "ExtraPackages": extraPackages,
            "BKStepID": stepID,
            "StepName": stepName,
            "StepVisible": stepVisible,
            "OutputFileTypes": outputVisibility,
        }

        self.bkSteps[stepIDInternal] = stepBKInfo
        self.__addBKPassStep()

        return name

    #############################################################################

    def _constructOutputFilesList(self, filesTypesList):
        """Build list of dictionary of output file types, including HIST case.

        :param list filesTypesList: a list of file types (str)
        :returns: a list with a dictionary of file types, lowered
        """

        outputList = []

        for fileType in filesTypesList:
            fileDict = {}
            fileDict["outputDataType"] = fileType.lower()
            outputList.append(fileDict)

        return outputList

    #############################################################################

    def __addBKPassStep(self):
        """Internal method to add BKK parameters."""
        bkPass = "BKProcessingPass"
        description = "BKProcessingPassInfo"
        self.LHCbJob._addParameter(
            self.LHCbJob.workflow, bkPass, "dict", self.bkSteps, description
        )  # pylint: disable=protected-access

    #############################################################################

    def addFinalizationStep(self, modulesList=None):
        """Add the finalization step to the workflow (some defaults are inserted)

        :param list modulesList: the list of modules names this step is made of. If None, a default is taken.
        :returns: None
        """
        if modulesList is None:
            modulesList = ["UploadOutputData", "UploadLogFile", "UploadMC", "FailoverRequest"]

        if "Job_Finalization" not in self.LHCbJob.workflow.step_definitions:
            jobFinalizationStepDef = getStepDefinition(
                "Job_Finalization", importLine="LHCbDIRAC.Workflow.Modules", modulesNameList=modulesList
            )
            self.LHCbJob.workflow.addStep(jobFinalizationStepDef)

        # create the step instance add it to the wf
        self.LHCbJob.workflow.createStepInstance("Job_Finalization", "finalization")

    #############################################################################

    def _lastParameters(self):
        """Add the last parameters before creating the xml file containing the
        workflow."""

        self.LHCbJob._addParameter(
            self.LHCbJob.workflow, "gaudiSteps", "list", self.gaudiSteps, "list of Gaudi Steps"
        )  # pylint: disable=protected-access
        self.LHCbJob._addParameter(
            self.LHCbJob.workflow, "outputSEs", "dict", self.outputSEs, "dictionary of output SEs"
        )  # pylint: disable=protected-access

    def __createWorkflow(self, name=""):
        """Create XML of the workflow."""
        self._lastParameters()

        if not name:
            name = self.LHCbJob.workflow.getName()
        if not re.search("xml$", name):
            name = f"{name}.xml"
        if os.path.exists(name):
            shutil.move(name, f"{name}.backup")
        name = name.replace("/", "").replace("\\", "")
        self.LHCbJob.workflow.toXMLFile(name)

        return name

    #############################################################################

    def runLocal(self):
        """Create XML workflow for local testing then reformulate as a job and run
        locally."""

        xmlFileName = self.__createWorkflow()
        # it makes a job (a Worklow, with Parameters), out of the xml file
        return LHCbJob(xmlFileName).runLocal()

    #############################################################################

    def __getProductionParameters(
        self, prodXMLFile, prodID, groupDescription="", bkPassInfo={}, derivedProd=0, reqID=0
    ):
        """This method will publish production parameters."""

        prodWorkflow = Workflow(prodXMLFile)

        parameters = {}
        info = []

        for parameterName in (
            "Priority",
            "CondDBTag",
            "DDDBTag",
            "DQTag",
            "eventType",
            "processingPass",
            "FractionToProcess",
            "MinFilesToProcess",
            "configName",
            "configVersion",
            "outputDataFileMask",
            "JobType",
            "MaxNumberOfTasks",
        ):
            try:
                parameters[parameterName] = prodWorkflow.findParameter(parameterName).getValue()
                info.append(f"{parameterName}: {prodWorkflow.findParameter(parameterName).getValue()}")
            except AttributeError:
                continue

        parameters["SizeGroup"] = self.jobFileGroupSize

        if prodWorkflow.findParameter("InputData"):  # now only comes from BK query
            prodWorkflow.findParameter("InputData").setValue("")
            gLogger.verbose("Resetting input data for production to null, this comes from a BK query...")
            prodXMLFile = self.__createWorkflow(prodXMLFile)
            # prodWorkflow.toXMLFile(prodXMLFile)

        if self.transformationFamily:
            parameters["TransformationFamily"] = self.transformationFamily

        if not bkPassInfo:
            bkPassInfo = prodWorkflow.findParameter("BKProcessingPass").getValue()
        if not groupDescription:
            groupDescription = prodWorkflow.findParameter("groupDescription").getValue()

        parameters["BKCondition"] = prodWorkflow.findParameter("conditions").getValue()
        parameters["BKProcessingPass"] = bkPassInfo
        parameters["groupDescription"] = groupDescription
        parameters["RequestID"] = reqID
        parameters["DerivedProduction"] = derivedProd

        result = self.__getOutputLFNs(prodID, "99999999", prodXMLFile)
        if not result["OK"]:
            gLogger.error("Could not create production LFNs", result)

        outputLFNs = result["Value"]
        parameters["OutputLFNs"] = outputLFNs

        # the list of output directories is later used for consistency check and for removing output data
        outputDirectories = []
        del outputLFNs["BookkeepingLFNs"]  # since ProductionOutputData uses the file mask
        for i in outputLFNs.values():
            for j in i:
                outputDir = f"{j.split(str(prodID))[0]}{prodID}"
                if outputDir not in outputDirectories:
                    outputDirectories.append(outputDir)

        parameters["OutputDirectories"] = outputDirectories

        # Now for the steps of the workflow
        stepKeys = sorted(bkPassInfo)
        for step in stepKeys:
            info.append(f"====> {bkPassInfo[step]['ApplicationName']} {bkPassInfo[step]['ApplicationVersion']} {step}")
            info.append(f"{bkPassInfo[step]['ApplicationName']} Option Files:")
            if optionsFiles := bkPassInfo[step]["OptionFiles"]:
                if optionsFiles.startswith("{"):
                    info.append(pprint.pformat(json.loads(optionsFiles)))
                else:
                    info.extend(optionsFiles.split(";"))
            if bkPassInfo[step]["ExtraPackages"]:
                info.append(f"ExtraPackages: {bkPassInfo[step]['ExtraPackages']}")

        try:
            if parameters["BkQuery"]:
                info.append("\nBK Input Data Query:")
                for bkn, bkv in parameters["BkQuery"].items():
                    info.append(f"{bkn}= {bkv}")
        except KeyError:
            pass

        # BK output directories (very useful)
        bkPaths = []
        bkOutputPath = "{}/{}/{}/{}/{}/{}".format(
            parameters["configName"],
            parameters["configVersion"],
            parameters["BKCondition"],
            parameters.get("processingPass", ""),
            parameters["groupDescription"],
            parameters["eventType"],
        )
        fileTypes = parameters["outputDataFileMask"]
        fileTypes = [a.upper() for a in fileTypes.split(";")]

        # Annoying that histograms are extension root
        if "ROOT" in fileTypes:
            fileTypes.remove("ROOT")
            fileTypes.append("HIST")

        for ft in fileTypes:
            bkPaths.append(f"{bkOutputPath}/{ft}")
        parameters["BKPaths"] = bkPaths
        info.append("\nBK Browsing Paths:\n%s" % ("\n".join(bkPaths)))
        infoString = "\n".join(info)
        parameters["DetailedInfo"] = infoString

        gLogger.verbose(f"Parameters that will be added: {parameters}")

        return parameters

    #############################################################################

    def create(self, publish=True, wfString="", requestID=0, reqUsed=0):
        """Will create the production and subsequently publish to the BK.
        Production parameters are also added at this point.

        publish = True - will add production to the production management system
                  False - does not publish the production

        The workflow XML is created regardless of the flags.
        """

        if wfString:
            self.LHCbJob.workflow = fromXMLString(wfString)
        #      self.name = self.LHCbJob.workflow.getName()

        bkConditions = self.LHCbJob.workflow.findParameter("conditions").getValue()

        bkSteps = self.LHCbJob.workflow.findParameter("BKProcessingPass").getValue()

        bkDictStep = {}

        # Add the BK conditions metadata / name
        simConds = self.bkkClient.getSimConditions()
        if not simConds["OK"]:
            gLogger.error(f"Could not retrieve conditions data from BK:\n{simConds}")
            return simConds
        simulationDescriptions = []
        for record in simConds["Value"]:
            simulationDescriptions.append(str(record[1]))

        if bkConditions not in simulationDescriptions:
            gLogger.verbose(f"Assuming BK conditions {bkConditions} are DataTakingConditions")
            bkDictStep["DataTakingConditions"] = bkConditions
        else:
            gLogger.verbose(f"Found simulation conditions for {bkConditions}")
            bkDictStep["SimulationConditions"] = bkConditions

        descShort = self.LHCbJob.workflow.getDescrShort()
        descLong = self.LHCbJob.workflow.getDescription()

        prodID = 0
        if publish:
            self.setParameter("ProcessingType", "JDL", str(self.prodGroup), "ProductionGroupOrType")
            self.setParameter("Priority", "JDL", str(self.priority), "UserPriority")

            try:
                fileName = self.__createWorkflow()
            except Exception as x:  # pylint: disable=broad-except
                gLogger.error(x)
                return S_ERROR("Could not create workflow")

            gLogger.verbose(f"Workflow XML file name is: {fileName}")

            workflowBody = ""
            if os.path.exists(fileName):
                with open(fileName) as fopen:
                    workflowBody = fopen.read()
            else:
                return S_ERROR("Could not get workflow body")

            # Standard parameters
            self.transformation.setTransformationName(fileName)
            self.transformation.setTransformationGroup(self.prodGroup)
            self.transformation.setDescription(descShort)
            self.transformation.setLongDescription(descLong)
            self.transformation.setType(self.LHCbJob.type)
            self.transformation.setBody(workflowBody)
            self.transformation.setPlugin(self.plugin)
            self.transformation.setBkQuery(self.inputBKSelection)
            self.transformation.setTransformationFamily(self.transformationFamily)
            self.transformation.setFileMask(self.inputFileMask)
            self.transformation.setGroupSize(int(self.jobFileGroupSize))
            self.transformation.setInheritedFrom(int(self.ancestorProduction))

            result = self.transformation.addTransformation()

            if not result["OK"]:
                gLogger.error(f"Problem creating production:\n{result}")
                return result
            prodID = result["Value"]
            gLogger.info(f"Production {prodID} successfully created")

            # All other parameters
            groupDesc = self.LHCbJob.workflow.findParameter("groupDescription").getValue()
            paramsDict = self.__getProductionParameters(
                prodID=prodID,
                prodXMLFile=fileName,
                groupDescription=groupDesc,
                bkPassInfo=bkSteps,
                reqID=requestID,
                derivedProd=self.ancestorProduction,
            )
            for parName, parValue in paramsDict.items():
                result = getattr(self.transformation, "set" + parName)(parValue)

        else:
            gLogger.verbose("Publish flag is disabled, using default production ID")

        bkDictStep["Production"] = int(prodID)

        if self.inputBKSelection:
            queryProdID = int(self.inputBKSelection.get("ProductionID", 0))
            queryProcPass = (
                self.inputBKSelection.get("ProcessingPass", "")
                if self.inputBKSelection.get("ProcessingPass", "") != "All"
                else ""
            )

            if queryProdID:
                inputPass = self.bkkClient.getProductionProcessingPass(queryProdID)
                if not inputPass["OK"]:
                    gLogger.error(inputPass)
                    gLogger.error(
                        "Production %s was created but BK processing pass for %d was not found" % (prodID, queryProdID)
                    )
                    return inputPass
                inputPass = inputPass["Value"]
                gLogger.info(
                    "Setting %d as BK input production for %s with processing pass %s"
                    % (queryProdID, prodID, inputPass)
                )
                bkDictStep["InputProductionTotalProcessingPass"] = inputPass
            elif queryProcPass:
                gLogger.info(
                    "Adding input BK processing pass for production %s from input data query: %s"
                    % (prodID, queryProcPass)
                )
                bkDictStep["InputProductionTotalProcessingPass"] = queryProcPass
        else:
            # has to account for an input processing pass anyway, if there is one,
            # or output files may not have the output processing pass correctly computed
            try:
                bkDictStep["InputProductionTotalProcessingPass"] = self.LHCbJob.workflow.findParameter(
                    "processingPass"
                ).getValue()
            except AttributeError:
                pass

        stepList = []
        stepKeys = sorted(bkSteps)
        # The BK needs an ordered list of steps
        for step in stepKeys:
            stepID = bkSteps[step]["BKStepID"]
            if stepID:
                stepName = bkSteps[step]["StepName"]
                stepVisible = bkSteps[step]["StepVisible"]
                outputFileTypes = bkSteps[step]["OutputFileTypes"]
                stepList.append(
                    {
                        "StepId": int(stepID),
                        "OutputFileTypes": outputFileTypes,
                        "StepName": stepName,
                        "Visible": stepVisible,
                    }
                )

        # This is the last component necessary for the BK publishing (post reorganisation)
        bkDictStep["Steps"] = stepList
        bkDictStep["EventType"] = paramsDict["eventType"]

        bkDictStep["ConfigName"] = self.LHCbJob.workflow.findParameter("configName").getValue()
        bkDictStep["ConfigVersion"] = self.LHCbJob.workflow.findParameter("configVersion").getValue()

        if publish:
            gLogger.verbose(f"Attempting to publish production {prodID} to the BK")
            result = self.bkkClient.addProduction(bkDictStep)
            if not result["OK"]:
                gLogger.error(result)
                return result

        if requestID and publish:
            reqDict = {"ProductionID": int(prodID), "RequestID": requestID, "Used": reqUsed, "BkEvents": 0}
            result = ProductionRequestClient(timeout=120).addProductionToRequest(reqDict)
            if not result["OK"]:
                gLogger.error(f"Attempt to add production {prodID} to request {requestID} failed: {result['Message']} ")
                gLogger.error(f"Dictionary below:\n{reqDict}")
            else:
                gLogger.info(
                    f"Successfully added production {prodID} to request {requestID} with flag set to {reqUsed}"
                )

        return S_OK(prodID)

    #############################################################################

    def __getOutputLFNs(self, prodID="12345", prodJobID="6789", prodXMLFile=""):
        """Will construct the output LFNs for the production for visual
        inspection."""
        if not prodXMLFile:
            gLogger.verbose("Using workflow object to generate XML file")
            prodXMLFile = self.__createWorkflow()

        job = LHCbJob(prodXMLFile)
        result = preSubmissionLFNs(
            job._getParameters(),
            job.workflow.createCode(),  # pylint: disable=protected-access
            productionID=prodID,
            jobID=prodJobID,
        )
        if not result["OK"]:
            return result
        lfns = result["Value"]
        gLogger.verbose(lfns)
        return result

    #############################################################################

    def setFileMask(self, fileMask="", stepMask=""):
        """Output data related parameters."""
        if fileMask:
            if isinstance(fileMask, list):
                fileMask = ";".join(fileMask)
            self.setParameter("outputDataFileMask", "string", fileMask, "outputDataFileMask")

        if stepMask:
            if isinstance(stepMask, list):
                stepMask = ";".join(stepMask)
            self.LHCbJob._addParameter(
                self.LHCbJob.workflow, "outputDataStep", "string", stepMask, "outputDataStep Mask"
            )  # pylint: disable=protected-access

    #############################################################################

    def banTier1s(self):
        """Sets Tier1s as banned."""
        tier1s = []
        sites = getSites()
        if not sites["OK"]:
            return sites

        for site in sites["Value"]:
            tier = getSiteTier(site)
            if not tier["OK"]:
                return tier
            if int(tier["Value"]) in (0, 1):
                tier1s.append(site)

        self.LHCbJob.setBannedSites(tier1s)

    #############################################################################

    def banSites(self, listOfSites):
        """Sets Sites as banned."""
        sitesToBan = []
        sites = getSites()
        if not sites["OK"]:
            return sites

        sites = sites["Value"]
        for site in listOfSites:
            if site in sites:
                sitesToBan.append(site)

        self.LHCbJob.setBannedSites(sitesToBan)

    #############################################################################

    def setOutputMode(self, outputMode):
        """Sets output mode for all jobs, this can be 'Local' or 'Any'."""
        if not outputMode.lower().capitalize() in ("Local", "Any", "Run"):
            raise TypeError("Output mode must be 'Local' or 'Any' or 'Run'")
        self.setParameter("outputMode", "string", outputMode.lower().capitalize(), "SEResolutionPolicy")

    #############################################################################

    def setBKParameters(self, configName, configVersion, groupDescriptionOrStepsList, conditions):
        """Sets BK parameters for production."""
        self.setParameter("configName", "string", configName, "ConfigName")
        self.setParameter("configVersion", "string", configVersion, "ConfigVersion")
        if isinstance(groupDescriptionOrStepsList, list):
            # in this case we assume it is a list of steps (stepsListDict in
            # ProductionRequest module), so we calculate the BK path
            groupDescription = ""
            for step in groupDescriptionOrStepsList:
                if step["Visible"] == "Y":
                    groupDescription = os.path.sep.join([groupDescription, step["ProcessingPass"]])
        else:
            groupDescription = groupDescriptionOrStepsList
        self.setParameter("groupDescription", "string", groupDescription, "GroupDescription")
        self.setParameter("conditions", "string", conditions, "SimOrDataTakingCondsString")
        self.setParameter("simDescription", "string", conditions, "SimDescription")

    #############################################################################
    # properties

    def set_transformationFamily(self, value):
        if isinstance(value, int):
            value = str(value)
        self._transformationFamily = value

    def get_transformationFamily(self):
        return self._transformationFamily

    startRun = property(get_transformationFamily, set_transformationFamily)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
:mod: StorageUsageDB

.. module: StorageUsageDB

:synopsis: StorageUsageDB class is a front-end to the Storage Usage Database.
"""
import os

# # from DIRAC
from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Base.DB import DB


def _standardDirectory(dirPath):
    """This adds a / at the end of the path if not there"""
    return os.path.join(dirPath, "")


#############################################################################


class StorageUsageDB(DB):
    """.. class:: StorageUsageDB"""

    def __init__(self, parentLogger=None):
        """Standard Constructor."""
        DB.__init__(self, "StorageUsageDB", "DataManagement/StorageUsageDB", parentLogger=parentLogger)
        self.__initializeDB()

    def __initializeDB(self):
        """Create the tables."""
        result = self._query("show tables")
        if not result["OK"]:
            return result
        tablesInDB = [t[0] for t in result["Value"]]
        tablesToCreate = {}
        self.__tablesDesc = {}
        self.__tablesDesc["su_Directory"] = {
            "Fields": {
                "DID": "INTEGER UNSIGNED AUTO_INCREMENT NOT NULL",
                "Path": "VARCHAR(255) NOT NULL",
                "Files": "INTEGER UNSIGNED NOT NULL",
                "Size": "BIGINT UNSIGNED NOT NULL",
            },
            "PrimaryKey": "Path",
            "Indexes": {"id": ["dID"]},
        }

        self.__tablesDesc["su_SEUsage"] = {
            "Fields": {
                "DID": "INTEGER UNSIGNED NOT NULL",
                "SEName": "VARCHAR(32) NOT NULL",
                "Files": "INTEGER UNSIGNED NOT NULL",
                "Size": "BIGINT UNSIGNED NOT NULL",
                "Updated": "DATETIME NOT NULL",
            },
            "PrimaryKey": ["DID", "SEName"],
            "Indexes": {"SE": ["SEName"]},
        }

        self.__tablesDesc["se_Usage"] = {
            "Fields": {
                "DID": "INTEGER UNSIGNED NOT NULL",
                "SEName": "VARCHAR(32) NOT NULL",
                "Files": "INTEGER UNSIGNED NOT NULL",
                "Size": "BIGINT UNSIGNED NOT NULL",
                "Updated": "DATETIME NOT NULL",
            },
            "PrimaryKey": ["DID", "SEName"],
        }
        self.__tablesDesc["se_STSummary"] = {
            "Fields": {
                "Site": "VARCHAR(32) NOT NULL",
                "SpaceToken": "VARCHAR(32) NOT NULL",
                "TotalSize": "BIGINT UNSIGNED NOT NULL",
                "TotalFiles": "INTEGER UNSIGNED NOT NULL",
                "Updated": "DATETIME NOT NULL",
            },
            "PrimaryKey": ["Site", "SpaceToken"],
        }
        self.__tablesDesc["problematicDirs"] = {
            "Fields": {
                "DID": "INTEGER UNSIGNED AUTO_INCREMENT NOT NULL",
                "Path": "VARCHAR(255) NOT NULL",
                "Site": "VARCHAR(32) NOT NULL",
                "SpaceToken": "VARCHAR(32) NOT NULL",
                "ReplicaType": "VARCHAR(32) NOT NULL",
                "SEFiles": "INTEGER UNSIGNED NOT NULL",
                "LFCFiles": "INTEGER UNSIGNED NOT NULL",
                "SESize": "BIGINT UNSIGNED NOT NULL",
                "LFCSize": "BIGINT UNSIGNED NOT NULL",
                "Problem": "VARCHAR(255) NOT NULL",
                "Updated": "DATETIME NOT NULL",
            },
            "PrimaryKey": ["DID"],
        }
        self.__tablesDesc["Popularity"] = {
            "Fields": {
                "ID": "INTEGER UNSIGNED AUTO_INCREMENT NOT NULL",
                "Path": "VARCHAR(255) NOT NULL",
                "Site": "VARCHAR(32) NOT NULL",
                "Count": "INTEGER UNSIGNED NOT NULL",
                "InsertTime": "DATETIME NOT NULL",
                "Status": "VARCHAR(32) NOT NULL",
            },
            "PrimaryKey": ["ID"],
            "Indexes": {"InsertTime": ["InsertTime"], "Status": ["Status"]},
        }
        self.__tablesDesc["DirMetadata"] = {
            "Fields": {
                "DID": "INTEGER UNSIGNED NOT NULL",
                "ConfigName": "VARCHAR(64) NOT NULL",
                "ConfigVersion": "VARCHAR(64) NOT NULL",
                "Conditions": "VARCHAR(128) NOT NULL",
                "ProcessingPass": "VARCHAR(255) NOT NULL",
                "EventType": "VARCHAR(255) NOT NULL",
                "FileType": "VARCHAR(255) NOT NULL",
                "Production": "VARCHAR(64) NOT NULL",
                "Visibility": "VARCHAR(4) NOT NULL",
            },
            "PrimaryKey": ["DID"],
        }
        # These are templates for a possible popularity summary table
        # self.__tablesDesc['BKDatasets'] = { 'Fields': { 'DSID' : 'INTEGER UNSIGNED AUTO_INCREMENT NOT NULL',
        #                                                'DSPath':'VARCHAR(255) NOT NULL',
        #                                                'CreationTime':'DATETIME NOT NULL',
        #                                                'Archived':'BOOL'},
        #                                    'PrimaryKey' : ['DSID'],
        #                                    'Indexes' : { 'DSPath' : 'DSPath'}
        #                                  }
        # self.__tablesDesc['DatasetUsage'] = { 'Fields' : { 'DSID': 'INTEGER UNSIGNED NOT NULL',
        #                                                  'Week' : 'INTEGER UNSIGNED NOT NULL',
        #                                                  'Usage' : 'INTEGER UNSIGNED NOT NULL',
        #                                                  'Files' : 'INTEGER UNSIGNED NOT NULL',
        #                                                  'Size' : 'BIGINT UNSIGNED NOT NULL',
        #                                                  },
        #                                     'Indexes': { 'Week':'Week',
        #                                                 'DSID':'DSID'}
        #                                     }

        for tableName in self.__tablesDesc:
            if tableName not in tablesInDB:
                tablesToCreate[tableName] = self.__tablesDesc[tableName]
        return self._createTables(tablesToCreate)

    def __sqlDirList(self, dirList):
        """change a list of directories to an escape one."""
        return [self._escapeString(_standardDirectory(dirPath))["Value"] for dirPath in dirList]

    def __stripDirectory(self, dirPath):
        """Remove trailing / in directory names."""
        return self._escapeString(os.path.realpath(dirPath))["Value"][1:-1]

    ################
    # Bulk insertion
    ################

    def publishDirectories(self, directoryDict):
        """Inserts a group of directories with their usage."""

        if not directoryDict:
            return S_OK()
        self.log.debug(f"in publishDirectories: directoryDict is: \n {str(directoryDict)}")
        result = self.__getIDs(directoryDict)
        if not result["OK"]:
            return result
        dirIDs = result["Value"]
        for path, pathInfo in directoryDict.items():
            path = _standardDirectory(path)
            try:
                files = int(pathInfo["Files"])
                size = int(pathInfo["TotalSize"])
            except ValueError as e:
                return S_ERROR(f"Values must be ints: {repr(e)}")
            seUsage = pathInfo["SEUsage"]
            sqlpath = self._escapeString(path)["Value"]
            if path in dirIDs:
                sqlCmd = "UPDATE `su_Directory` SET Files=%d, Size=%d WHERE Path = %s" % (files, size, sqlpath)
                result = self._update(sqlCmd)
                if not result["OK"]:
                    self.log.error("Cannot update directory", f"{path}: {result['Message']}")
                    return result
            else:
                sqlCmd = "INSERT INTO `su_Directory` (DID, Path, Files, Size) VALUES ( 0, %s, %d, %d )" % (
                    sqlpath,
                    files,
                    size,
                )
                result = self._update(sqlCmd)
                if not result["OK"]:
                    self.log.error("Cannot insert directory", f"{path}: {result['Message']}")
                    continue
                dirIDs[path] = result["lastRowId"]
            result = self.__updateSEUsage(dirIDs[path], seUsage, path)
            if not result["OK"]:
                self.log.error("Cannot update SE usage", f"{path}:{seUsage}: {result['Message']}")
                return result
        return S_OK()

    ####
    # Insert/update to problematicDirs
    ####
    def publishToProblematicDirs(self, directoryDict):
        """Publish an entry into the problematic data directory."""
        for path, pathInfo in directoryDict.items():
            path = _standardDirectory(path)
            spaceToken = pathInfo["SpaceToken"]
            problem = pathInfo["Problem"]
            site = pathInfo["Site"]
            replicaType = pathInfo["ReplicaType"]
            try:
                seFiles = int(pathInfo["Files"])
                seSize = int(pathInfo["Size"])
                lfcFiles = int(pathInfo["LFCFiles"])
                lfcSize = int(pathInfo["LFCSize"])
            except ValueError:
                return S_ERROR("ERROR: Files and Size have to be integer!")
            # check if the tuple (path,Site,SpaceToken) already exists in the table
            sqlPath = self._escapeString(path)["Value"]
            sqlSpaceToken = self._escapeString(spaceToken)["Value"]
            sqlProblem = self._escapeString(problem)["Value"]
            sqlSite = self._escapeString(site)["Value"]
            sqlReplicaType = self._escapeString(replicaType)["Value"]
            sqlCmd = "SELECT DID FROM problematicDirs WHERE Path={} AND Site={} AND SpaceToken={}".format(
                sqlPath,
                sqlSite,
                sqlSpaceToken,
            )
            self.log.verbose(f"in publishToProblematicDirs, sqlCmd: {sqlCmd}")
            result = self._query(sqlCmd)
            if not result["OK"]:
                self.log.error("Failed to query problematicDirs")
                return result
            did = result["Value"]
            if did:
                # there is an entry for (path, Site, SpaceToken), make an update of the row
                sqlCmd = (
                    "UPDATE `problematicDirs` SET SEFiles=%d, SESize=%d, LFCFiles=%d, "
                    "LFCSize=%d, Problem=%s, ReplicaType=%s, Updated=UTC_TIMESTAMP() WHERE "
                    "Path = %s and Site = %s and SpaceToken=%s"
                    % (seFiles, seSize, lfcFiles, lfcSize, sqlProblem, sqlReplicaType, sqlPath, sqlSite, sqlSpaceToken)
                )
                self.log.verbose(f"in publishToProblematicDirs, sqlCmd: {sqlCmd}")
                result = self._update(sqlCmd)
                if not result["OK"]:
                    self.log.error(f"Cannot update row ({path}, {site}) in problematicDirs")
                    return result
            else:
                # entry is not there, make an insert of a new row
                sqlCmd = (
                    "INSERT INTO problematicDirs (Path, Site, SpaceToken, SEFiles, SESize, "
                    "LFCFiles, LFCSize, Problem, ReplicaType, Updated) VALUES "
                    "( %s, %s, %s, %d, %d, %d, %d, %s, %s, UTC_TIMESTAMP())"
                    % (sqlPath, sqlSite, sqlSpaceToken, seFiles, seSize, lfcFiles, lfcSize, sqlProblem, sqlReplicaType)
                )
                self.log.verbose(f"in publishToProblematicDirs, sqlCmd: {sqlCmd}")
                result = self._update(sqlCmd)
                if not result["OK"]:
                    self.log.error(f"Cannot insert row ({path}, {site}) in problematicDirs")
                    return result
        return S_OK()

    # ##
    # Insert/Update to se_Usage table
    # ##

    def publishToSEReplicas(self, directoryDict):
        """Publish an entry to se_Usage table."""
        for path, pathInfo in directoryDict.items():
            path = _standardDirectory(path)
            seName = pathInfo["SEName"]
            try:
                files = int(pathInfo["Files"])
                size = int(pathInfo["Size"])
            except ValueError:
                return S_ERROR("ERROR: Files and Size have to be integer!")
            # check if the tuple (path,SE) already exists in the table
            sqlPath = self._escapeString(path)["Value"]
            sqlSeName = self._escapeString(seName)["Value"]
            sqlCmd = (
                "SELECT d.DID FROM su_Directory AS d, se_Usage AS u WHERE "
                "d.DID=u.DID AND d.Path=%s AND u.SEName=%s" % (sqlPath, sqlSeName)
            )
            self.log.verbose(f"in publishToEReplias, sqlCmd: {sqlCmd}")
            result = self._query(sqlCmd)
            if not result["OK"]:
                self.log.error("Failed to query se_Usage")
                return result
            sqlRes = result["Value"]
            if sqlRes:
                try:
                    did = int(sqlRes[0][0])
                except ValueError:
                    return S_ERROR("ERROR: DID should be integer!")
                # there is an entry for (path, SEname), make an update of the row
                sqlCmd = (
                    "UPDATE `se_Usage` SET Files=%d, Size=%d, Updated=UTC_TIMESTAMP() WHERE "
                    "DID=%d AND SEName=%s" % (files, size, did, sqlSeName)
                )
                self.log.verbose(f"in publishToEReplias, sqlCmd: {sqlCmd}")
                result = self._update(sqlCmd)
                if not result["OK"]:
                    self.log.error(f"Cannot update row ({path}, {seName}) in se_Usage")
                    return result
            else:
                # entry is not there, make an insert of a new row
                # get the DID corresponding to this Path, from su_Directory table:
                result = self.__getIDs(directoryDict)
                if not result["OK"]:
                    self.log.error(f"Cannot getIds for directory {path}")
                    return result
                for dirKey in result["Value"]:
                    try:
                        did = int(result["Value"][dirKey])
                    except ValueError:
                        return S_ERROR("ERROR: DID should be integer!")
                sqlCmd = (
                    "INSERT INTO se_Usage (DID, SEName, Files, Size, Updated) VALUES "
                    " ( %d, %s, %d, %d, UTC_TIMESTAMP())" % (did, sqlSeName, files, size)
                )
                self.log.verbose(f"in publishToEReplias, sqlCmd: {sqlCmd}")
                result = self._update(sqlCmd)
                if not result["OK"]:
                    self.log.error(f"Cannot insert row ({path}, {seName}) in se_Usage")
                    return result
        return S_OK()

    def getIDs(self, dirList):
        """get IDs for list of directories."""
        return self.__getIDs(dirList)

    def __getIDs(self, dirList):
        """get IDs for list of directories."""
        dl = self.__sqlDirList(dirList)
        sqlCmd = f"SELECT Path, DID FROM `su_Directory` WHERE Path in ( {', '.join(dl)} )"
        self.log.verbose(f"in __getIDs, query: {sqlCmd}")
        result = self._query(sqlCmd)
        if not result["OK"]:
            return result
        return S_OK(dict(result["Value"]))

    def __getIDsFromProblematicDirs(self, dirList):
        """Get the ID from the problematicDirs table, for a given directory
        {Path:seInfo}"""
        self.log.verbose(f"entry to be removed: {dirList}")
        dirPath, dirInfo = list(dirList.items())[0]
        dirPath = _standardDirectory(dirPath)
        spaceToken = dirInfo["SpaceToken"]
        sqlPath = self._escapeString(dirPath)["Value"]
        sqlSpaceToken = self._escapeString(spaceToken)["Value"]
        site = dirInfo["Site"]
        sqlSite = self._escapeString(site)["Value"]
        sqlCmd = "SELECT Path, DID FROM problematicDirs WHERE Path={} AND Site={} and SpaceToken={}".format(
            sqlPath,
            sqlSite,
            sqlSpaceToken,
        )
        self.log.verbose(f"in __getIDsFromProblematicDirs, query: {sqlCmd}")
        result = self._query(sqlCmd)
        if not result["OK"]:
            return result
        return S_OK(dict(result["Value"]))

    def __getIDsFromSe_Usage(self, dirList):
        """Get the ID of the entry corresponding to the directory {path:seInfo}"""
        dirPath, dirInfo = list(dirList.items())[0]
        dirPath = _standardDirectory(dirPath)

        # take into account that SEName might not be available
        se = ""
        site = ""
        try:
            se = dirInfo["SEName"]
        except KeyError:
            self.log.info("SEName attribute not available!")
        if not se:
            self.log.info("The SEName attribute is not available. Try to get the Site attribute..")
            try:
                site = dirInfo["Site"]
            except KeyError:
                self.log.info("Site attribute not available!")
        if se:
            sqlSE = self._escapeString(se)["Value"]
        else:
            sqlSE = ""
        if site:
            sqlSite = self._escapeString(site)["Value"]
        else:
            sqlSite = ""
        sqlPath = self._escapeString(dirPath)["Value"]
        if sqlSE:
            sqlCmd = (
                "SELECT d.Path, d.DID FROM se_Usage r, su_Directory d WHERE d.DID=r.DID \
      AND d.Path=%s AND r.SEName=%s"
                % (sqlPath, sqlSE)
            )
        elif sqlSite:
            sqlCmd = (
                "SELECT d.Path, d.DID FROM se_Usage r, su_Directory d WHERE d.DID=r.DID \
      AND d.Path=%s AND r.SEName LIKE '%s%%'"
                % (sqlPath, sqlSite)
            )
        else:
            self.log.error("Not enough information to formulate the query: either Site or SEName should be provided ")
            return S_ERROR()
        self.log.verbose(f"in __getIDsFromSe_Usage, query: {sqlCmd}")
        result = self._query(sqlCmd)
        if not result["OK"]:
            return result
        return S_OK(dict(result["Value"]))

    def __updateSEUsage(self, dirID, seUsage, path):
        """Update the SE usage table."""
        sqlCmd = "DELETE FROM `su_SEUsage` WHERE DID=%d" % dirID
        result = self._update(sqlCmd)
        if not result["OK"]:
            return result
        if not seUsage:
            # This indicates there are files without replicas in the FC
            self.log.warn("Ooops. Dir has no SEUsage!", f"ID {dirID}, path {path}")
            return S_OK()
        sqlVals = []
        # Insert data
        for seName in seUsage:
            try:
                size = seUsage[seName]["Size"]
                files = seUsage[seName]["Files"]
                # Protect against hiccoughs of the DFC that may return negative numbers
                if size < 0 or files < 0:
                    self.log.error("SEUsage < 0 for %s @ %s: %d files, %d size" % (path, seName, files, size))
                    continue
                # This line should be in the try otherwise the exception will never work
                sqlVals.append(
                    "( %d, %s, %d, %d, UTC_TIMESTAMP() )" % (dirID, self._escapeString(seName)["Value"], size, files)
                )
            except ValueError as e:
                return S_ERROR(f"Values must be ints: {repr(e)}")
        # It may happen that no SE is OK
        if not sqlVals:
            return S_OK()
        sqlIn = f"INSERT INTO `su_SEUsage` ( DID, SEname, Size, Files, Updated ) VALUES {', '.join(sqlVals)}"
        return self._update(sqlIn)

    #############
    # Remove dir
    #############

    def __getIDsLike(self, dirPath):
        dirPath = self.__stripDirectory(dirPath)
        sqlCmd = f"SELECT Path, DID FROM `su_Directory` WHERE Path LIKE '{dirPath}/%'"
        result = self._query(sqlCmd)
        if not result["OK"]:
            return result
        return S_OK(dict(result["Value"]))

    def removeDirectory(self, dirPath):
        result = self.__getIDsLike(dirPath)
        if not result["OK"]:
            return result
        dirIDs = result["Value"]
        if not dirIDs:
            return S_OK()
        sqlIDs = ", ".join([str(dirIDs[path]) for path in dirIDs])
        for tableName in ("su_Directory", "su_SEUsage"):
            sqlCmd = f"DELETE FROM `{tableName}` WHERE DID in ( {sqlIDs} )"
            result = self._update(sqlCmd)
            if not result["OK"]:
                return result
        return S_OK()

    def removeDirFromSe_Usage(self, dirDict):
        """Remove the entry corresponding to the tuple (path, SE) from the se_Usage
        table.

        This function is typically called when a directory is found to be a
        problematic directory, and before inserting it into the
        problematicDirs table, it is necessary to remove it from the
        se_Usage table, if it exists there. In general, one same directory
        can only exist into either the se_Usage or the problematicDirs
        table.
        """
        deletedDirs = 0
        result = self.__getIDsFromSe_Usage(dirDict)
        if not result["OK"]:
            return result
        dirIDs = result["Value"]
        if not dirIDs:
            return S_OK(deletedDirs)
        deletedDirs = len(dirIDs)
        if deletedDirs > 1:
            return S_ERROR("There should not be more than 1 match for a tuple (path,SE)!!")
        sqlIDs = ", ".join([str(dirIDs[path]) for path in dirIDs])
        sqlCmd = f"DELETE FROM se_Usage WHERE DID in ( {sqlIDs} )"
        result = self._update(sqlCmd)
        if not result["OK"]:
            return result
        return S_OK(deletedDirs)

    def removeDirFromProblematicDirs(self, dirDict):
        """Remove an entry from the problematicDirs table.

        This is typically used when a directory is found to be correctly
        registered in the FC and it was previously inserted in the
        problematic data table: before inserting the entry into the se_Usage
        table, it has to be removed from the problematicDirs
        """
        deletedDirs = 0
        result = self.__getIDsFromProblematicDirs(dirDict)
        if not result["OK"]:
            return result
        self.log.verbose(f"sqlCmd result for __getIDsFromProblematicDirs is: {result}")
        dirIDs = result["Value"]
        if not dirIDs:
            self.log.info("No directory to remove")
            return S_OK(deletedDirs)
        deletedDirs = len(dirIDs)
        if deletedDirs > 1:
            return S_ERROR("There should not be more than 1 match for a tuple (path,SE)!!")
        self.log.info(f"remove directories: {dirIDs}")
        sqlIDs = ", ".join([str(dirIDs[path]) for path in dirIDs])
        sqlCmd = f"DELETE FROM problematicDirs WHERE DID in ( {sqlIDs} )"
        result = self._update(sqlCmd)
        if not result["OK"]:
            return result
        self.log.verbose(f"sqlCmd result for DELETE is: {result}")
        return S_OK(deletedDirs)

    def removeAllFromProblematicDirs(self, site=False):
        """Remove all entries from the problematicDirs table, for a given site
        (optional)."""
        if site:
            sqlSite = self._escapeString(site)["Value"]
            sqlCmd = f"DELETE FROM problematicDirs WHERE Site={sqlSite}"
        else:
            sqlCmd = "DELETE FROM problematicDirs"
        self.log.verbose(f"in removeAllFromProblematicDirs, command: {sqlCmd}")
        result = self._update(sqlCmd)
        if not result["OK"]:
            return result
        self.log.verbose(f"sqlCmd result for DELETE is: {result}")
        deletedDirs = result["Value"]
        return S_OK(deletedDirs)

    ################
    # Clean outdated entries
    ################

    def purgeOutdatedEntries(self, rootDir=False, outdatedSeconds=86400, preserveDirsList=None):
        preserveDirsList = preserveDirsList if preserveDirsList else []
        try:
            outdatedSeconds = max(1, int(outdatedSeconds))
        except ValueError:
            return S_ERROR("Ooutdated seconds needs to be a number")

        sqlCond = ["TIMESTAMPDIFF( SECOND, su.Updated, UTC_TIMESTAMP() ) > %d" % outdatedSeconds]
        sqlTables = ["`su_SEUsage` as su"]
        sqlLimit = 1000
        if rootDir:
            rootDir = self.__stripDirectory(rootDir)
            sqlTables.append("`su_Directory` as d")
            sqlCond.append(f"d.Path LIKE '{rootDir}/%'")
            sqlCond.append("d.DID = su.DID")
        if preserveDirsList:
            for ignoreDir in preserveDirsList:
                ignoreDir = self.__stripDirectory(ignoreDir)
                sqlCond.append(f"d.Path NOT LIKE '{ignoreDir}/%'")
        sqlCmd = "SELECT DISTINCT su.DID FROM %s WHERE %s LIMIT %d" % (
            ", ".join(sqlTables),
            " AND ".join(sqlCond),
            sqlLimit,
        )
        cleaned = 0
        while True:
            result = self._query(sqlCmd)
            if not result["OK"]:
                return result
            idList = [str(row[0]) for row in result["Value"]]
            if not idList:
                break
            cleaned += len(idList)
            sqlIdList = ", ".join(idList)
            for tblName in ("su_Directory", "su_SEUsage"):
                result = self._update(f"DELETE FROM `{tblName}` WHERE DID IN ( {sqlIdList} )")
                if not result["OK"]:
                    return result
            if len(idList) < sqlLimit:
                break
        return S_OK(cleaned)

    ###############
    #  Get Storage info
    ###############

    def getStorageElementSelection(self):
        """Retireve the possible selections available through the web-monitor."""
        sqlCmd = "SELECT DISTINCT SEName FROM `su_SEUsage` ORDER BY SEName"
        result = self._query(sqlCmd)
        if not result["OK"]:
            return result
        return S_OK([row[0] for row in result["Value"]])

    def __getStorageCond(self, path, fileType, production, SEs):
        path = self.__stripDirectory(path)
        sqlCond = ["d.DID = su.DID", f"d.Path LIKE '{path}/%'"]
        if fileType:
            fileType = self._escapeString(fileType)["Value"][1:-1]
            sqlCond.append(f"d.Path LIKE '%/{fileType}%'")
        if production:
            try:
                sqlCond.append("d.Path LIKE '%%/%08.d%%'" % int(production))
            except ValueError:
                return S_ERROR("production has to be a number")
        if SEs:
            sqlCond.append(f"su.SEName in ( {', '.join([self._escapeString(SEName)['Value'] for SEName in SEs])} )")
        return sqlCond

    def __getStorageSummary(self, path, fileType=False, production=False, SEs=None, groupingField="su.SEName"):
        """Retrieves the storage summary for all of the known directories."""
        if SEs is None:
            SEs = []
        sqlCond = self.__getStorageCond(path, fileType, production, SEs)
        sqlFields = (groupingField, "SUM(su.Size)", "SUM(su.Files)")
        sqlCmd = "SELECT {} FROM `su_SEUsage` as su,`su_Directory` as d WHERE {} GROUP BY {}".format(
            ", ".join(sqlFields),
            " AND ".join(sqlCond),
            groupingField,
        )
        result = self._query(sqlCmd)
        if not result["OK"]:
            return result
        usageDict = {}
        for gf, size, files in result["Value"]:
            usageDict[gf] = {"Size": int(size), "Files": int(files)}
        return S_OK(usageDict)

    def getStorageSummary(self, path, fileType=False, production=False, SEs=[]):
        return self.__getStorageSummary(path, fileType, production, SEs, "su.SEName")

    def getStorageDirectorySummary(self, path, fileType=False, production=False, SEs=[]):
        return self.__getStorageSummary(path, fileType, production, SEs, "d.Path")

    def getStorageDirectories(self, path, fileType=False, production=False, SEs=[]):
        sqlCond = self.__getStorageCond(path, fileType, production, SEs)
        sqlCmd = f"SELECT DISTINCT d.Path FROM `su_SEUsage` as su,`su_Directory` as d WHERE {' AND '.join(sqlCond)}"
        result = self._query(sqlCmd)
        if not result["OK"]:
            return result
        return S_OK([row[0] for row in result["Value"]])

    @staticmethod
    def __userExpression(fieldName="d.Path"):
        return f"SUBSTRING_INDEX( SUBSTRING_INDEX( {fieldName}, '/', 5 ), '/', -1 )"

    def getUserStorageUsage(self, userName=False):
        """Get the usage in the SEs."""
        sqlCond = ["d.DID = su.DID"]
        if userName:
            userName = self._escapeString(userName)["Value"][1:-1]
            sqlCond.append(f"d.Path LIKE '/lhcb/user/{userName[0]}/{userName}/%")
        else:
            sqlCond.append("d.Path LIKE '/lhcb/user/_/%'")
        sqlCmd = "SELECT %s, SUM( su.Size ) FROM `su_Directory` AS d, `su_SEUsage` AS su " "WHERE %s GROUP BY %s" % (
            self.__userExpression(),
            " AND ".join(sqlCond),
            self.__userExpression(),
        )
        result = self._query(sqlCmd)
        if not result["OK"]:
            return result
        userStorage = {}
        for row in result["Value"]:
            userStorage[row[0]] = int(row[1])
        return S_OK(userStorage)

    def getUserSummaryPerSE(self, userName=False):
        sqlUser = self.__userExpression()
        sqlFields = (sqlUser, "su.SEName", "SUM(su.Size)", "SUM(su.Files)")
        sqlCond = ["d.DID = su.DID", "d.Path LIKE '/lhcb/user/_/%'"]
        if userName:
            sqlCond.append(f"{sqlUser} = {self._escapeString(userName)['Value']}")
        sqlGroup = (sqlUser, "su.SEName")
        sqlCmd = "SELECT {} FROM `su_Directory` AS d, `su_SEUsage` AS su WHERE {} GROUP BY {}".format(
            ", ".join(sqlFields),
            " AND ".join(sqlCond),
            ", ".join(sqlGroup),
        )
        result = self._query(sqlCmd)
        if not result["OK"]:
            return result
        userData = {}
        for row in result["Value"]:
            userName = row[0]
            seName = row[1]
            if userName not in userData:
                userData[userName] = {}
            userData[userName][seName] = {"Size": int(row[2]), "Files": int(row[3])}
        return S_OK(userData)

    def getDirectorySummaryPerSE(self, directory):
        """Queries the DB and get a summary (total size and files) for the given
        directory."""
        sqlCmd = (
            "SELECT su.SEName, SUM(su.Size), SUM(su.Files) FROM su_Directory AS d, su_SEUsage AS su "
            " WHERE d.DID=su.DID AND d.Path LIKE '%s%%' GROUP BY su.SEName" % (directory)
        )
        self.log.verbose(f"in getDirectorySummaryPerSE, sqlCmd: {sqlCmd} ")
        result = self._query(sqlCmd)
        if not result["OK"]:
            return result
        data = {}
        for row in result["Value"]:
            seName = row[0]
            if seName not in data:
                data[seName] = {}
            data[seName] = {"Size": int(row[1]), "Files": int(row[2])}
        return S_OK(data)

    def publishTose_STSummary(self, site, spaceToken, totalSize, totalFiles, storageDumpLastUpdate):
        """Publish total size and total files extracted from the storage dumps to
        the se_STSummary."""
        try:
            sqlTotalSize = int(totalSize)
            sqlTotalFiles = int(totalFiles)
        except ValueError as e:
            return S_ERROR(f"Values must be ints: {repr(e)}")
        sqlSpaceToken = self._escapeString(spaceToken)["Value"]
        sqlSite = self._escapeString(site)["Value"]
        sqlSDUpdate = self._escapeString(storageDumpLastUpdate)["Value"]
        # check if there is already an entry for the tuple (site, spaceTok)
        res = self.__getSTSummary(site, spaceToken)
        self.log.verbose(f" __getSTSummary returned: {res} ")
        if not res["OK"]:
            return S_ERROR(res)
        if not res["Value"]:
            self.log.warn(f"Entry for site={site}, spaceToken={spaceToken} is not there => insert new entry ")
            sqlCmd = (
                "INSERT INTO `se_STSummary` (Site, SpaceToken, TotalSize, TotalFiles, Updated) VALUES "
                "( %s, %s, %d, %d, %s)" % (sqlSite, sqlSpaceToken, sqlTotalSize, sqlTotalFiles, sqlSDUpdate)
            )

        else:
            self.log.verbose(f"Entry for site={site}, spaceToken={spaceToken} is there => update entry ")
            sqlCmd = (
                "UPDATE `se_STSummary` SET TotalSize=%d, TotalFiles=%d, Updated=%s WHERE "
                "Site=%s and SpaceToken=%s" % (sqlTotalSize, sqlTotalFiles, sqlSDUpdate, sqlSite, sqlSpaceToken)
            )

        self.log.verbose(f"in publishTose_STSummary, sqlCmd: {sqlCmd} ")
        result = self._update(sqlCmd)
        if not result["OK"]:
            self.log.error("Cannot insert entry", f"{sqlSite},{sqlSpaceToken}: {result['Message']}")
            return result
        return S_OK()

    def __getSTSummary(self, site, spaceToken=False):
        """Get total files and total size for the space token identified by the
        input arguments (site, space token) , if space token is not specified,
        return all entries relative to the site."""

        sqlSite = self._escapeString(site)["Value"]
        sqlCond = [f"Site={sqlSite}"]
        if spaceToken:
            sqlSpaceToken = self._escapeString(spaceToken)["Value"]
            sqlCond.append(f"SpaceToken={sqlSpaceToken}")

        sqlCmd = "SELECT Site, SpaceToken, TotalSize, TotalFiles, Updated FROM `se_STSummary` WHERE %s" % " AND ".join(
            sqlCond
        )
        self.log.verbose(f"in __getSTSummary sqlCmd is: {sqlCmd} ")
        result = self._query(sqlCmd)
        self.log.verbose(f" __getSTSummary result: {result} ")
        if not result["OK"]:
            return S_ERROR(result)
        return S_OK(result["Value"])

    def getSTSummary(self, site, spaceToken=False):
        """Returns a summary of space usage for the given site, on the basis of the
        information provided by the SRM db dumps from sites."""
        return self.__getSTSummary(site, spaceToken)

    def removeSTSummary(self, site, spaceToken=False):
        """Remove from se_STSummary table the entry relative to the given site and
        space token (if specified).

        If no space token is specified, remove all entries relative to site
        """
        sqlSite = self._escapeString(site)["Value"]
        sqlCond = [f"Site={sqlSite}"]
        if spaceToken:
            sqlSpaceToken = self._escapeString(spaceToken)["Value"]
            sqlCond.append(f"SpaceToken={sqlSpaceToken}")

        sqlCmd = f"DELETE FROM `se_STSummary` WHERE {' AND '.join(sqlCond)}"
        self.log.verbose(f"in __getSTSummary, sqlCmd: {sqlCmd} ")
        result = self._update(sqlCmd)
        self.log.verbose(f"in __getSTSummary result: {result} ")
        if not result["OK"]:
            return S_ERROR(result)
        return S_OK(result["Value"])

    def getProblematicDirsSummary(self, site, problem=False):
        """Get a summary of problematic directories for a given site and given
        problem (optional)"""
        sqlSite = self._escapeString(site)["Value"]
        sqlCond = [f"Site={sqlSite}"]
        if problem:
            sqlProblem = self._escapeString(problem)["Value"]
            sqlCond.append(f"Problem={sqlProblem}")

        sqlCmd = (
            "SELECT Site, SpaceToken, LFCFiles, SEFiles, Path, Problem, ReplicaType FROM `problematicDirs` "
            "WHERE %s" % (" AND ".join(sqlCond))
        )
        self.log.verbose(f"in getProblematicDirsSummary sqlCmd: {sqlCmd} ")
        result = self._query(sqlCmd)
        self.log.verbose(f"in getProblematicDirsSummary result: {result} ")
        if not result["OK"]:
            return S_ERROR(result["Message"])
        return S_OK(result["Value"])

    def getRunSummaryPerSE(self, run):
        """Queries the DB and get a summary (total size and files) per SE  for the
        given run. It assumes that the path in the LFC where the run's file are
        stored is like:

        /lhcb/data/[YEAR]/RAW/[STREAM]/[PARTITION]/[ACTIVITY]/[RUNNO]/
        """
        # try and implement bulk query
        # check the type of run
        data = {}
        if isinstance(run, list):
            for thisRun in run:
                sqlCmd = (
                    "SELECT su.SEName, SUM(su.Size), SUM(su.Files)  FROM su_Directory AS d, su_SEUsage AS su WHERE "
                    "d.DID=su.DID AND d.Path LIKE '/lhcb/data/%%/RAW/%%/%%/%%/%d/' GROUP BY su.SEName" % (thisRun)
                )
                self.log.verbose(f"in getRunSummaryPerSE, sqlCmd: {sqlCmd} ")
                result = self._query(sqlCmd)
                if not result["OK"]:
                    return S_ERROR(result)
                data[thisRun] = {}
                for row in result["Value"]:
                    seName = row[0]
                    data[thisRun][seName] = {"Size": int(row[1]), "Files": int(row[2])}
        else:
            sqlCmd = (
                "SELECT su.SEName, SUM(su.Size), SUM(su.Files)  FROM su_Directory AS d, su_SEUsage AS su WHERE "
                "d.DID=su.DID and d.Path LIKE '/lhcb/data/%%/RAW/%%/%%/%%/%d/' GROUP BY su.SEName" % (run)
            )
            self.log.verbose(f"in getRunSummaryPerSE, sqlCmd: {sqlCmd} ")
            result = self._query(sqlCmd)
            if not result["OK"]:
                return S_ERROR(result)
            for row in result["Value"]:
                seName = row[0]
                data[seName] = {"Size": int(row[1]), "Files": int(row[2])}

        return S_OK(data)

    def __getAllReplicasInFC(self, path):
        """Queries the su_seUsage table to get all the entries relative to a given
        path registered in the FC.

        Returns for every replica the SE, the update, the files and the
        total size
        """
        path = _standardDirectory(path)
        sqlCmd = f"SELECT DID FROM su_Directory where Path like '{path}%'"
        result = self._query(sqlCmd)
        if not result["OK"]:
            return result
        if not result["Value"]:  # no replica found
            return S_OK(result)

        for row in result["Value"]:
            did = row[0]
            # warning! this has to be corrected! there is no need to make a loop! the query should return only one row
            # add a check for that. And return an error if it return more than one row

        sqlCmd = f"SELECT Files, Updated, SEName, Size FROM su_SEUsage WHERE DID={did}"
        result = self._query(sqlCmd)
        if not result["OK"]:
            return result
        replicasData = {}
        replicasData[path] = {}
        for row in result["Value"]:
            seName = row[2]
            if seName in replicasData[path]:
                return S_ERROR("There cannot be two replicas on the same SE!")
            replicasData[path][seName] = {}
            replicasData[path][seName]["Files"] = row[0]
            replicasData[path][seName]["Updated"] = row[1]
            replicasData[path][seName]["Size"] = row[3]

        return S_OK(replicasData)

    def getAllReplicasInFC(self, path):
        return self.__getAllReplicasInFC(path)

    ######
    # Catalog queries
    ######

    def __getCatalogCond(self, path, fileType, production):
        path = self.__stripDirectory(path)
        sqlCond = [f"Path LIKE '{path}/%'"]
        if fileType:
            fileType = self._escapeString(fileType)["Value"][1:-1]
            sqlCond.append(f"Path LIKE '%/{fileType}%'")
        if production:
            try:
                sqlCond.append("Path LIKE '%%/%08.d%%'" % int(production))
            except ValueError:
                return S_ERROR("production has to be a number")
        return sqlCond

    def __getSummary(self, path, fileType=False, production=False, groupingField="Path"):
        """Retrieves the storage summary for all of the known directories."""
        sqlCond = self.__getCatalogCond(path, fileType, production)
        sqlFields = (groupingField, "SUM(Size)", "SUM(Files)")
        sqlCmd = "SELECT {} FROM `su_Directory` WHERE {} GROUP BY {}".format(
            ", ".join(sqlFields),
            " AND ".join(sqlCond),
            groupingField,
        )
        result = self._query(sqlCmd)
        if not result["OK"]:
            return result
        usageDict = {}
        for gf, size, files in result["Value"]:
            usageDict[gf] = {"Size": int(size), "Files": int(files)}
        return S_OK(usageDict)

    def getSummary(self, path, fileType=False, production=False):
        return self.__getSummary(path, fileType, production, "Path")

    def getUserSummary(self, userName=False):
        if userName:
            userName = self._escapeString(userName)["Value"][1:-1]
            path = f"/lhcb/user/{userName[0]}/{userName}"
        else:
            path = "/lhcb/user/_/%"
        return self.__getSummary(path, groupingField=self.__userExpression("Path"))

    ######
    # methods to interact with Popularity table
    ######

    def sendDataUsageReport(self, site, directoryDict, status="New"):
        """Add a new trace in the Popularity table."""
        self.log.verbose(f"in addPopCount: dirDict: {directoryDict} site: {site}")
        if not directoryDict:
            return S_OK()
        sqlSite = self._escapeString(site)["Value"]
        insertedEntries = 0
        for di, count in directoryDict.items():
            di = _standardDirectory(di)
            sqlPath = self._escapeString(di)["Value"]
            sqlStatus = self._escapeString(status)["Value"]
            if not isinstance(count, int):
                self.log.warn(f"in sendDataUsageReport: type is not correct {count}")
                continue
            sqlCmd = (
                "INSERT INTO `Popularity` ( Path, Site, Count, Status, InsertTime) VALUES "
                "( %s, %s, %d, %s, UTC_TIMESTAMP() )" % (sqlPath, sqlSite, count, sqlStatus)
            )
            self.log.verbose(f"in sendDataUsageReport, sqlCmd: {sqlCmd} ")
            result = self._update(sqlCmd)
            if not result["OK"]:
                self.log.error(f"Cannot insert entry: {result['Message']}")
                continue
            insertedEntries += 1
        return S_OK(insertedEntries)

    def sendDataUsageReport_2(self, directoryDict):
        """Add a new trace in the Popularity table (new version which takes in
        input a dictionary per trace) Each trace corresponds to a directory, and
        the mandatory keys are the site and the count.

        Optional keys are the status and the creation time
        """
        self.log.verbose(f"in addPopCount: dirDict: {directoryDict}")
        if not directoryDict:
            return S_OK()
        insertedEntries = 0
        for di in directoryDict:
            try:
                site = directoryDict[di]["site"]
                count = directoryDict[di]["count"]
            except KeyError:
                self.log.error(f"input directoryDict should specify site and count keys . {str(directoryDict)} ")
                return S_ERROR("wrong arguments format")
            # set default value for status
            status = directoryDict[di].get("status", "New")

            di = _standardDirectory(di)
            sqlPath = self._escapeString(di)["Value"]
            sqlStatus = self._escapeString(status)["Value"]
            sqlSite = self._escapeString(site)["Value"]
            if not isinstance(count, int):
                self.log.warn(f"in sendDataUsageReport: type is not correct {count}")
                continue
            # by default, insert the record with the current time, unless the input directoryDict specifies the creation time
            insertTime = directoryDict[di].get("creationTime")
            sqlInsertTime = self._escapeString(insertTime)["Value"] if insertTime else "UTC_TIMESTAMP()"
            sqlCmd = (
                "INSERT INTO `Popularity` ( Path, Site, Count, Status, InsertTime) VALUES "
                "( %s, %s, %d, %s, %s )" % (sqlPath, sqlSite, count, sqlStatus, sqlInsertTime)
            )
            self.log.verbose(f"in sendDataUsageReport_2, sqlCmd: {sqlCmd} ")
            result = self._update(sqlCmd)
            if not result["OK"]:
                self.log.error(f"Cannot insert entry: {result['Message']}")
                continue
            insertedEntries += 1
        return S_OK(insertedEntries)

    def getDataUsageSummary(self, startTime, endTime, status="New"):
        """returns a summary of the counts for each tuple (Site,Path) in the given
        time interval."""
        if startTime > endTime:
            return S_OK()
        if not isinstance(startTime, str) or not isinstance(endTime, str):
            return S_ERROR("wrong arguments format")

        sqlStartTime = self._escapeString(startTime)["Value"]
        sqlEndTime = self._escapeString(endTime)["Value"]
        sqlStatus = self._escapeString(status)["Value"]
        sqlCmd = (
            "SELECT ID, Path, Site, Count, InsertTime from `Popularity` WHERE "
            "Status = %s AND InsertTime > %s AND InsertTime < %s " % (sqlStatus, sqlStartTime, sqlEndTime)
        )
        self.log.verbose(f"in getDataUsageSummary, sqlCmd: {sqlCmd} ")
        return self._query(sqlCmd)

    def getDataUsageForDirectory(self, path):
        """returns all entries for a given path."""
        sqlCmd = (
            "SELECT ID, Path, Site, Count, InsertTime from `Popularity` WHERE Path = %s"
            % self._escapeString(path)["Value"]
        )
        return self._query(sqlCmd)

    def updatePopEntryStatus(self, idList, newStatus):
        """Update the status of the entry identified by the IDList, to the status
        specified by Newstatus."""
        if not idList:
            self.log.info("updatePopEntryStatus: no entry to be updated")
            return S_OK()

        sqlStatus = self._escapeString(newStatus)["Value"]
        strIdList = [str(iD) for iD in idList]
        sqlIdList = ", ".join(strIdList)
        self.log.verbose(f"updatePopEntryStatus: sqlIdList {sqlIdList} ")
        sqlCmd = f"UPDATE `Popularity` SET Status={sqlStatus} WHERE ID IN ( {sqlIdList} )"
        result = self._update(sqlCmd)
        if not result["OK"]:
            return result
        return result

    def insertToDirMetadata(self, directoryDict):
        """Inserts a new entry into the DirMetadata table.

        The input dictionary should contain, for each lfn directory, a
        dictionary with all the necessary metadata provided by the
        bookkeeping
        """

        if not directoryDict:
            return S_OK()
        # get the IDs from the su_Directory table:
        result = self.__getIDs(directoryDict)
        if not result["OK"]:
            return result
        dirIDs = result["Value"]
        if not dirIDs:
            errMsg = f"ERROR! the directory {directoryDict} was not found in the LFN directory table! "
            self.log.error(f"{errMsg}")
            return S_ERROR(errMsg)
        self.log.verbose(f"in insertToDirMetadata: dirIDs: {dirIDs}")
        # returns a dictionary of type:
        #  {dir1: ID1, dir2: ID2, ...}
        insertedEntries = 0
        for dirKey in directoryDict:
            if dirKey not in dirIDs:
                self.log.warn(f"in insertToDirMetadata: found no DID for directory {dirKey}")
                continue
            iD = int(dirIDs[dirKey])
            try:
                info = directoryDict[dirKey]
                sqlConfigname = self._escapeString(info["ConfigName"])["Value"]
                sqlConfigversion = self._escapeString(info["ConfigVersion"])["Value"]
                sqlConditions = self._escapeString(info["ConditionDescription"])["Value"]
                sqlProcpass = self._escapeString(info["ProcessingPass"])["Value"]
                sqlEvttype = self._escapeString(info["EventType"])["Value"]
                sqlFiletype = self._escapeString(info["FileType"])["Value"]
                sqlProd = self._escapeString(info["Production"])["Value"]
                sqlVisi = self._escapeString(info["Visibility"])["Value"]
            except KeyError:
                self.log.error(f"in insertToDirMetadata: the input dict was not correctly formatted: {directoryDict}")
                return S_ERROR(f"Key error in input dictionary {directoryDict}")
            sqlCmd = (
                "INSERT INTO `DirMetadata` ( DID, ConfigName, ConfigVersion, Conditions, ProcessingPass, "
                "EventType, FileType, Production, Visibility ) VALUES ( %d, %s, %s, %s, %s, %s, %s, %s, %s )"
                % (
                    iD,
                    sqlConfigname,
                    sqlConfigversion,
                    sqlConditions,
                    sqlProcpass,
                    sqlEvttype,
                    sqlFiletype,
                    sqlProd,
                    sqlVisi,
                )
            )
            self.log.verbose(f"in insertToDirMetadata, sqlCmd: {sqlCmd} ")
            result = self._update(sqlCmd)
            if not result["OK"]:
                self.log.error(f"Cannot insert entry: {result['Message']}")
                return S_ERROR(result["Message"])
            insertedEntries += 1
        return S_OK(insertedEntries)

    def getDirMetadata(self, dirList):
        """Return the directory meta-data, which have been previously provided by
        the Bookkeeping and stored in the DirMetadata table.

        The input is a list of LFN directories
        """
        if not dirList:
            return S_OK()
        # get the IDs from the su_Directory table:
        stdDirList = [_standardDirectory(dirPath) for dirPath in dirList]
        result = self.__getIDs(stdDirList)
        if not result["OK"]:
            return S_ERROR(result["Message"])
        dirIDs = result["Value"]
        if not dirIDs:
            errMsg = f"ERROR! the directories {str(stdDirList)} were not found in the LFN directory table! "
            self.log.error(f"{errMsg}")
            return S_ERROR(errMsg)
        self.log.verbose(f"in getDirMetadata, dirIDs: {dirIDs}")
        # returns a dictionary of type:
        #  {dir1: ID1, dir2: ID2, ...}
        idList = []
        for dirKey in stdDirList:
            if dirKey not in dirIDs:
                self.log.warn(f"in getDirMetadata: found no DID for directory {dirKey}")
            else:
                idList.append(str(dirIDs[dirKey]))
        sqlCmd = (
            "SELECT DID, ConfigName, ConfigVersion, Conditions, ProcessingPass, EventType, FileType, "
            "Production, Visibility FROM `DirMetadata` WHERE DID in ( %s )"
            % ", ".join(["'%s'" % did for did in idList])
        )
        self.log.verbose(f"in getDirMetadata, sqlCmd: {sqlCmd} ")
        result = self._query(sqlCmd)
        if not result["OK"]:
            return S_ERROR(result["Message"])
        # We must build the list of tuples with the same order than in the input
        metadata = {}
        resList = result["Value"]
        for dirPath, dirKey in zip(dirList, stdDirList):
            metadata[dirPath] = ()
            if dirKey in dirIDs:
                did = dirIDs[dirKey]
                for t in resList:
                    if t[0] == did:
                        metadata[dirPath] = t
                        break
        self.log.verbose(f"getDirMetadata returns {str(metadata)}")
        return S_OK(metadata)


################################################################################



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Gets a list of files from BK and checks if they have common ancestors."""
import time
from DIRAC import gLogger
from DIRAC.Core.Base.Script import Script

fixIt = None


def removeFile(lfns, transClient, dm, bkClient):
    """Method for removing a file in the RM as well as in the Transformation
    system."""
    res = bkClient.setFilesInvisible(lfns)
    if res["OK"]:
        gLogger.always("Files were made invisible in BK")
    else:
        gLogger.always("Error making files invisible in BK", res["Message"])
    gLogger.always(f"Removing {len(lfns)} files from disk")
    res = dm.getReplicas(lfns)
    success = res.get("Value", {}).get("Successful", [])
    failed = res.get("Value", {}).get("Failed", lfns)
    if res["OK"] and success:
        seLfns = {}
        removeFiles = list(success)
        for lfn in success:
            for se in success[lfn]:
                if not se.endswith("ARCHIVE"):
                    seLfns.setdefault(se, []).append(lfn)
                else:
                    # If there is an archive, we cannot remove the file completely
                    if lfn in removeFiles:
                        removeFiles.remove(lfn)
        if removeFiles:
            gLogger.always(f"Removing completely {len(removeFiles)} files")
            res = dm.removeFile(removeFiles)
            if not res["OK"]:
                gLogger.fatal(f"Error removing {len(removeFiles)} files", res["Message"])
            else:
                success = res["Value"]["Successful"]
                failed = res["Value"]["Failed"]
                gLogger.always(f"Successfully fully removed {len(success)} files that only had the duplicate ancestors")
                if failed:
                    errors = {}
                    for lfn, reason in failed.items():
                        errors.setdefault(reason, []).append(lfn)
                    gLogger.error(
                        f"Failed to remove {len(failed)} files",
                        "\n".join(f"{reason}: {errors[reason]}" for reason in errors),
                    )
        success = []
        failed = {}
        for se, replicas in seLfns.items():
            replicas = [lfn for lfn in replicas if lfn not in removeFiles]
            if replicas:
                res = dm.removeReplica(se, replicas)
                if not res["OK"]:
                    gLogger.fatal("Error removing replicas", res["Message"])
                else:
                    success += res["Value"]["Successful"]
                    if res["Value"]["Failed"]:
                        failed[se] = res["Value"]["Failed"]
        if success:
            gLogger.always(
                f"Successfully removed {len(success)} replicas of files that only had the duplicate ancestors"
            )
        if failed:
            errors = {}
            for se in failed:
                for lfn, reason in failed[se].items():
                    errors.setdefault(reason + f" @{se}", []).append(lfn)
            gLogger.error("Failed to remove replicas\n", "\n".join(f"{reason}: {errors[reason]}" for reason in errors))
        elif not success and not removeFiles:
            gLogger.always("None of the files had replicas left on disk...")
        if not success:
            return
        res = transClient.getTransformationFiles({"LFN": success})
        if res["OK"]:
            transFiles = {}
            processedFiles = {}
            for fileDict in res["Value"]:
                if fileDict["Status"] == "Processed":
                    processedFiles.setdefault(fileDict["TransformationID"], []).append(fileDict["LFN"])
                else:
                    transFiles.setdefault(fileDict["TransformationID"], []).append(fileDict["LFN"])
            if processedFiles:
                gLogger.always("WARNING: some files were already processed!")
                for transID, lfns in processedFiles.items():
                    gLogger.always("%d: %d files" % (transID, len(lfns)))
                    gLogger.info("\n".join(sorted(lfns)))
            for transID, lfns in transFiles.items():
                res = transClient.setFileStatusForTransformation(transID, "Removed", lfns)
                if res["OK"]:
                    gLogger.always("%d: %d files set Removed" % (transID, len(lfns)))
                    gLogger.info("\n".join(sorted(lfns)))
                else:
                    gLogger.fatal(
                        "Error setting %d files to Removed status in transformation %d" % (len(lfns), transID)
                    )
    elif failed:
        gLogger.fatal(
            f"Failed to get replicas for {len(failed)} files:\n",
            res.get("Message", "\n".join(sorted(set(failed.values if isinstance(failed, type({})) else [])))),
        )
        gLogger.info(
            "\n".join(
                (f"{lfn}: {failed[lfn]}" for lfn in sorted(failed)) if isinstance(failed, dict) else sorted(failed)
            )
        )


def analyzeAncestors(commonAncestors, ancestors, transClient, dm, bkClient):
    """Analyse the list of common ancestors and checks whether one can remove
    some files."""
    lfnsToRemove = set()
    allLfns = [lfn for lfnStr in commonAncestors for lfn in lfnStr.split(",")]
    res = bkClient.getFileMetadata(allLfns)
    lfnRuns = {}
    if res["OK"]:
        for lfn, metadata in res["Value"]["Successful"].items():
            lfnRuns[lfn] = metadata["RunNumber"]
    for lfnStr, anc in commonAncestors.items():
        lfns = lfnStr.split(",")
        run = lfnRuns.get(lfns[0], "Unknown")
        gLogger.always("\n{} (run {}):\n\t{}".format("\n".join(lfns), str(run), "\n\t".join(anc)))
        # Now check if one of the files only has those common ancestors
        for lfn in sorted(lfns, reverse=True):
            if anc == ancestors[lfn]:
                if not fixIt:
                    gLogger.always(f"{lfn} only has those as ancestors. Use --FixIt to remove the file")
                else:
                    gLogger.always(f"{lfn} only has those as ancestors. We shall remove the file")
                    lfnsToRemove.add(lfn)
                    lfns.remove(lfn)
                    if len(lfns) == 1:
                        break
    if lfnsToRemove:
        removeFile(list(lfnsToRemove), transClient=transClient, dm=dm, bkClient=bkClient)


@Script()
def main():
    global fixIt

    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

    dmScript = DMScript()
    dmScript.registerFileSwitches()
    dmScript.registerBKSwitches()
    depth = 10
    Script.registerSwitch("", "Depth=", "Depth to which one should check ancestors (default %d)" % depth)
    Script.registerSwitch("", "FixIt", "Remove the files that only have the common ancestors")
    Script.registerSwitch("", "Verbose", "Set script level to INFO")

    Script.setUsageMessage(
        "\n".join(
            [
                __doc__,
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile]",
            ]
        )
    )
    Script.parseCommandLine(ignoreErrors=True)

    fixIt = False
    for switch in Script.getUnprocessedSwitches():
        if switch[0] == "Verbose":
            gLogger.setLevel("INFO")
        elif switch[0] == "FixIt":
            fixIt = True
        elif switch[0] == "Depth":
            depth = int(switch[1])

    # In case the user asked for specific LFNs
    lfnList = dmScript.getOption("LFNs", [])

    from LHCbDIRAC.DataManagementSystem.Client.ConsistencyChecks import ConsistencyChecks
    from DIRAC.DataManagementSystem.Client.DataManager import DataManager
    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
    from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

    dm = DataManager()
    bkClient = BookkeepingClient()
    transClient = TransformationClient()

    cc = ConsistencyChecks(transClient=transClient, dm=dm, bkClient=bkClient)
    cc.bkQuery = dmScript.getBKQuery()
    cc.lfns = lfnList
    cc.ancestorsDepth = depth
    startTime = time.time()
    cc.checkAncestors()

    # Print out the results
    gLogger.always(f"\nResults ({time.time() - startTime:.1f} seconds):")

    if not cc.commonAncestors:
        gLogger.always("No files found with common ancestors ==> OK")
    else:
        gLogger.always(f"Found {len(cc.commonAncestors)} sets of files with common ancestors ==> ERROR")
        analyzeAncestors(cc.commonAncestors, cc.ancestors, transClient=transClient, dm=dm, bkClient=bkClient)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

import DIRAC
from DIRAC.Core.Base.Script import Script


def usage():
    """usage.

    Prints script usage
    """
    print(f"Usage: {Script.scriptName} <Production ID> [<DIRAC Status>]")
    DIRAC.exit(2)


@Script()
def main():
    Script.parseCommandLine(ignoreErrors=True)

    from LHCbDIRAC.Interfaces.API.DiracProduction import DiracProduction

    args = Script.getPositionalArgs()

    if len(args) < 1:
        usage()

    diracProd = DiracProduction()
    prodID = args[0]

    stat = None
    if len(args) == 2:
        stat = args[1]

    result = diracProd.getProductionJobSummary(prodID, status=stat, printOutput=True)
    if result["OK"]:
        DIRAC.exit(0)
    elif "Message" in result:
        print(f"Getting production job summary failed with message:\n{result['Message']}")
        DIRAC.exit(2)
    else:
        print("Null result for getProductionJobSummary() call")
        DIRAC.exit(2)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""This script adds a user directory to the DFC or changes a user directory's
ownership."""
import os
import sys
from time import time

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    from DIRAC import exit, gLogger

    Script.registerSwitch("", "User=", "  User name (no default, mandatory)")
    Script.registerSwitch("", "Recursive", "  Set ownership recursively for existing user")

    Script.parseCommandLine()

    user = None
    recursive = False
    group = "lhcb_user"
    for switch in Script.getUnprocessedSwitches():
        if switch[0] == "User":
            user = switch[1].lower()
        elif switch[0] == "Recursive":
            recursive = True

    from DIRAC.Core.Security.ProxyInfo import getProxyInfo
    from LHCbDIRAC.DataManagementSystem.Utilities.FCUtilities import chown

    res = getProxyInfo()
    if not res["OK"]:
        gLogger.fatal("Can't get proxy info", res["Message"])
        exit(1)
    properties = res["Value"].get("groupProperties", [])
    if "FileCatalogManagement" not in properties:
        gLogger.error("You need to use a proxy from a group with FileCatalogManagement")
        exit(5)

    from DIRAC.Resources.Catalog.FileCatalogClient import FileCatalogClient

    dfc = FileCatalogClient()

    if user is None:
        Script.showHelp(exitCode=1)

    exists = False
    initial = user[0]
    baseDir = os.path.join("/lhcb", "user", initial, user)
    subDirectories = []
    if dfc.isDirectory(baseDir).get("Value", {}).get("Successful", {}).get(baseDir):
        gLogger.always("Directory already existing", baseDir)
        exists = True
        res = dfc.listDirectory(baseDir)
        if res["OK"]:
            success = res["Value"]["Successful"][baseDir]
            subDirectories = success["SubDirs"]
            if success.get("SubDirs") or success.get("Files"):
                gLogger.always(
                    "Directory is not empty:",
                    f" {len(success['Files'])} files / {len(success['SubDirs'])} subdirectories",
                )
            elif recursive:
                gLogger.always("Empty directory, recursive is useless...")
                recursive = False
    if not exists:
        gLogger.always("Creating directory", baseDir)
        res = dfc.createDirectory(baseDir)
        if not res["OK"]:
            gLogger.fatal("Error creating directory", res["Message"])
            exit(2)
    else:
        gLogger.always(f"Change{' recursively'} ownership of directory" if recursive else "", baseDir)
    res = chown(baseDir, user, group="lhcb_user", mode=0o755, recursive=False, fcClient=dfc)
    if not res["OK"]:
        gLogger.fatal("Error changing directory owner", res["Message"])
        exit(2)
    startTime = time()
    if recursive and subDirectories:
        res = chown(subDirectories, user, group="lhcb_user", mode=0o755, recursive=True, ndirs=1, fcClient=dfc)
        if not res["OK"]:
            gLogger.fatal("Error changing directory owner", res["Message"])
            exit(2)
        gLogger.always(
            "Successfully changed owner in %d directories in %.1f seconds" % (res["Value"], time() - startTime)
        )
    else:
        gLogger.always(f"Successfully changed owner in directory {baseDir} in {time() - startTime:.1f} seconds")

    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript
    from LHCbDIRAC.DataManagementSystem.Client.ScriptExecutors import executeLfnMetadata

    dmScript = DMScript()
    dmScript.setLFNs(baseDir)
    sys.stdout.write("Directory metadata: ")
    sys.stdout.flush()
    executeLfnMetadata(dmScript)
    exit(0)


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Whatever:

.. code-block:: python

  in_dict = {
    'EventTypeId': 93000000,
    'ConfigVersion': 'Collision10',
    'ProcessingPass': '/Real Data',
    'ConfigName': 'LHCb',
    'ConditionDescription': 'Beam3500GeV-VeloClosed-MagDown',
    'Production':7421
  }
"""
import tempfile

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Base.Client import Client, createClient
from DIRAC.Core.Tornado.Client.ClientSelector import TransferClientSelector as TransferClient
from DIRAC.Core.Utilities.Decorators import deprecated
from DIRAC.Core.Utilities.JEncode import strToIntDict

from LHCbDIRAC.BookkeepingSystem.Client import JEncoder
from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequestClient import ProductionRequestClient
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient


@createClient("Bookkeeping/BookkeepingManager")
class BookkeepingClient(Client):
    """This class expose the methods of the Bookkeeping Service."""

    def __init__(self, **kwargs):
        """c'tor.

        :param str url: can specify a specific URL
        """
        super().__init__(**kwargs)
        self.setServer("Bookkeeping/BookkeepingManager")
        self.timeout = 3600
        self.log = gLogger.getSubLogger("BookkeepingClient")

    #############################################################################
    def getAvailableFileTypes(self):
        """It returns all the available files which are registered to the bkk."""
        retVal = self._getRPC().getAvailableFileTypes()
        if retVal["OK"]:
            records = []
            parameters = ["FileType", "Description"]
            for record in retVal["Value"]:
                records += [list(record)]
            return S_OK({"ParameterNames": parameters, "Records": records, "TotalRecords": len(records)})
        return retVal

    #############################################################################
    def getFilesWithMetadata(self, in_dict):
        """It is used for retrieving a files with meta data for a given condition.

        :param dict in_dict: It can contains the following conditions:
          ``ConfigName``, ``ConfigVersion``, ``ConditionDescription``,
          ``EventType``, ``ProcessingPass``,``Production``,``RunNumber``,
          ``FileType``, ``DataQuality``, ``StartDate`` and ``EndDate``
        :return: files with meta data associated
        """
        in_dict = dict(in_dict)
        bkk = TransferClient(self.serverURL)
        in_dict["MethodName"] = "getFilesWithMetadata"
        params = JEncoder.dumps(in_dict)
        with tempfile.NamedTemporaryFile() as fileName:
            retVal = bkk.receiveFile(fileName.name, params)
            if not retVal["OK"]:
                return retVal
            value = JEncoder.load(open(fileName.name))
        return value

    #############################################################################
    def bulkJobInfo(self, in_dict):
        """It returns the job metadata information for a given condition:

        -a list of lfns
        - a list of DIRAC job ids
        - a list of jobNames

        :param dict in_dict: dictionary which has the following format: in_dict = {'lfn':[],jobId:[],jobName:[]}
        :return: job meta data
        """
        conditions = {}
        if isinstance(in_dict, str):
            conditions["lfn"] = in_dict.split(";")
        elif isinstance(in_dict, list):
            conditions["lfn"] = in_dict
        else:
            conditions = in_dict

        return self._getRPC().bulkJobInfo(conditions)

    #############################################################################
    def setFileDataQuality(self, lfns, flag):
        """It is used to set the files data quality.

        :param list lfns: list of LFNs or an LFN
        :param str flag: data quality flag: OK, UNCHECKED, etc.
        """
        if isinstance(lfns, str):
            lfns = lfns.split(";")

        return self._getRPC().setFileDataQuality(lfns, flag)

    #############################################################################
    def getFileAncestors(self, lfns, depth=0, replica=True):
        """Retrieve file ancestors.

        :param list lfns: list of LFNs
        :param int depth: depth of the processing chane
        :param bool replica: take into account the replica flag.

        :returns: It returns the ancestors of a file with metadata or a list of files
        """
        if isinstance(lfns, str):
            lfns = lfns.split(";")

        return self._getRPC().getFileAncestors(lfns, depth, replica)

    #############################################################################
    @deprecated("Use getFileDescendents")
    def getFileDescendants(self, lfns, depth=0, production=0, checkreplica=False):
        return self.getFileDescendents(lfns, depth, production, checkreplica)

    def getFileDescendents(self, lfns, depth=0, production=0, checkreplica=False):
        """Retrieve the file descendants.

        :param list lfns: list of LFNs
        :param int depth: depth of the processing chane
        :param bool replica: take into account the replica flag.
        :return: descendants of a file or a list of files.
        """
        if isinstance(lfns, str):
            lfns = lfns.split(";")

        return self._getRPC().getFileDescendents(lfns, depth, production, checkreplica)

    #############################################################################
    def addFiles(self, lfns):
        """It sets the replica flag Yes for a given list of files.

        :param list lfns: list of LFNs
        """
        if isinstance(lfns, str):
            lfns = lfns.split(";")
        return self._getRPC().addFiles(lfns)

    #############################################################################
    def removeFiles(self, lfns):
        """It removes the replica flag for a given list of files.

        :param list lfns: list of lfns
        """
        if isinstance(lfns, str):
            lfns = lfns.split(";")
        return self._getRPC().removeFiles(lfns)

    #############################################################################
    def getFileMetadata(self, lfns):
        """Retrieve the metadata information for a given file or a list of files.

        :param list lfns: list of LFNs
        :return: file metadata
        """
        if isinstance(lfns, str):
            lfns = lfns.split(";")
        return self._getRPC().getFileMetadata(lfns)

    #############################################################################
    def exists(self, lfns):
        """It used to check the existence of a list of files in the Bookkeeping
        Metadata catalogue.

        :param list lfns: list of LFNs
        :return: a dictionary with the lfns {'lfn':True/False}
        """
        if isinstance(lfns, str):
            lfns = lfns.split(";")
        return self._getRPC().exists(lfns)

    #############################################################################
    def getRunInformation(self, in_dict):
        """It returns run information and statistics.

        :param dict in_dict: contains a given conditions
        :return: run statistics Number of events, luminosity, etc.
        """
        if "Fields" not in in_dict:
            in_dict["Fields"] = [
                "ConfigName",
                "ConfigVersion",
                "JobStart",
                "JobEnd",
                "TCK",
                "FillNumber",
                "ProcessingPass",
                "ConditionDescription",
                "CONDDB",
                "DDDB",
            ]
        if "Statistics" in in_dict and not in_dict["Statistics"]:
            in_dict["Statistics"] = [
                "NbOfFiles",
                "EventStat",
                "FileSize",
                "FullStat",
                "Luminosity",
                "InstLumonosity",
                "EventType",
            ]

        return self._getRPC().getRunInformation(in_dict)

    #############################################################################
    def getRunInformations(self, runnb):
        """It returns run information and statistics.

        :param (int, str) runnb: run number
        :return: run statistics
        """
        # The service expects a int
        try:
            return self._getRPC().getRunInformations(int(runnb))
        except (ValueError, TypeError) as e:
            return S_ERROR(f"Invalid run number: {repr(e)}")

    #############################################################################
    def getRunFilesDataQuality(self, runs):
        """For retrieve the data quality for runs or files.

        :param list runs: list of run numbers.
        :return: run or file data quality
        """
        if isinstance(runs, str):
            runs = runs.split(";")
        elif isinstance(runs, int):
            runs = [runs]
        return self._getRPC().getRunFilesDataQuality(runs)

    #############################################################################
    def getRunExtendedDQOK(self, runnb):
        """For retrieve ExtendedDQOK for run.

        :param (int, str) runnb: run number.
        :return: list of system names
        """
        try:
            return self._getRPC().getRunExtendedDQOK(int(runnb))
        except (ValueError, TypeError) as e:
            return S_ERROR(f"Invalid run number: {repr(e)}")

    #############################################################################
    def setFilesInvisible(self, lfns):
        """It is used to set the file(s) invisible in the database.

        :paran list lfns: an lfn or list of lfns
        """
        if isinstance(lfns, str):
            lfns = lfns.split(";")
        return self._getRPC().setFilesInvisible(lfns)

    #############################################################################
    def setFilesVisible(self, lfns):
        """It is used to set the file(s) invisible in the database.

        :param list lfns: an lfn or list of lfns
        """
        if isinstance(lfns, str):
            lfns = lfns.split(";")
        return self._getRPC().setFilesVisible(lfns)

    #############################################################################
    def getFileTypeVersion(self, lfns):
        """For retrieving the file type version.

        :param list lfns: list of lfns
        :return: file type version
        """
        if isinstance(lfns, str):
            lfns = lfns.split(";")
        return self._getRPC().getFileTypeVersion(lfns)

    #############################################################################
    def getDirectoryMetadata(self, lfns):
        """For retrieving  meta data information for a given directory.

        :param list lfns: list of lfns
        :return: directory metadata
        """
        if isinstance(lfns, str):
            lfns = lfns.split(";")
        return self._getRPC().getDirectoryMetadata(lfns)

    #############################################################################
    def getRunsForFill(self, fillid):
        """For retrieving a list of runs.

        :param int fillid: fill number
        :return: runs for a given fill
        """
        try:
            fill = int(fillid)
        except ValueError as ex:
            return S_ERROR(ex)
        return self._getRPC().getRunsForFill(fill)

    #############################################################################
    def deleteSimulationConditions(self, simid):
        """It deletes a given simulation condition.

        :param int simid: simulation condition identifier
        """
        try:
            simid = int(simid)
        except ValueError as ex:
            return S_ERROR(ex)
        return self._getRPC().deleteSimulationConditions(simid)

    #############################################################################
    def getJobInputOutputFiles(self, diracjobids):
        """For retrieving input/output for a given Dirac job.

        :param list diracjobids: list of dirac job ids.
        :return: input/output file(s) for a given dirac job
        """
        if isinstance(diracjobids, int):
            diracjobids = [diracjobids]
        return self._getRPC().getJobInputOutputFiles(diracjobids)

    def fixRunLuminosity(self, runnumbers):
        """For fixing the luminosity for a given run(s)

        :param list runnumbers: list of run numbers.
        """
        if isinstance(runnumbers, int):
            runnumbers = [runnumbers]
        return self._getRPC().fixRunLuminosity(runnumbers)

    # The following method names are changed in the Bookkeeping client.

    #############################################################################
    def getFiles(self, in_dict):
        """It returns a list of files for a given condition.

        :param dict in_dict: contains a given conditions
        :return: list of files
        """
        in_dict = dict(in_dict)
        bkk = TransferClient(self.serverURL)
        in_dict["MethodName"] = "getFiles"
        params = JEncoder.dumps(in_dict)
        with tempfile.NamedTemporaryFile() as fileName:
            retVal = bkk.receiveFile(fileName.name, params)
            if not retVal["OK"]:
                return retVal
            return JEncoder.load(open(fileName.name))

    def getRunStatus(self, runs):
        """For retrieving the run status.

        :param list runs: list of runs
        :return: run status (finished, or not finished)
        """
        runnumbers = []
        if isinstance(runs, str):
            runnumbers = [int(run) for run in runs.split(";")]
        elif isinstance(runs, int):
            runnumbers += [runs]
        else:
            runnumbers = runs
        retVal = self._getRPC().getRunStatus(runnumbers)
        if retVal["OK"]:
            retVal["Value"]["Successful"] = strToIntDict(retVal["Value"]["Successful"])
        return retVal

    def getProcessingPass(self, in_dict, path=None):
        """This method is used to recursively browse the processing pass.

        :param dict in_dict: contains a given conditions: ConfigName', 'ConfigVersion', 'ConditionDescription',
          'Production','RunNumber', 'EventType'
        :param str path: To start the browsing you have to define the path as a root: path = '/'
        :return: processing pass for a given conditions. Note: it returns a list with two dictionary. First dictionary
          contains the processing passes while the second dictionary contains the event types.
        """
        if path is None:
            path = "/"
        return self._getRPC().getProcessingPass(in_dict, path)

    def getProductionFilesStatus(self, productionid=None, lfns=None):
        """Status of the files, which belong to a given production.

        :param str/int prodID: production (transformation) ID
        :param list lfns: list of LFNs
        :returns: the file status in the bkk for a given production or a list of lfns.
        """
        if lfns is None:
            lfns = []
        return self._getRPC().getProductionFilesStatus(productionid, lfns)

    def getProductionInformation(self, prodID):
        """Get the production information.

        :param str/int prodID: production (transformation) ID
        :returns: S_OK with dictionary of production info
        """
        res = self._getRPC().getProductionInformation(prodID)
        if not res["OK"]:
            return res

        prodInfo = res["Value"]

        res = self.getSteps(prodID)
        if not res["OK"]:
            return res

        prodInfo["Steps"] = res["Value"]

        return S_OK(prodInfo)

    def getSteps(self, prodID):
        """Fully resolve the steps of the production in input

        :param str/int prodID: production (transformation) ID
        :returns: S_OK with list of resolved steps
        """
        res = self._getRPC().getSteps(prodID)
        if not res["OK"]:
            return res
        steps = res["Value"]  # this is an ordered list
        if not steps:
            self.log.error(f"Production {prodID} does not have recorded steps")
            return S_ERROR("No recorded steps")
        # now we check if the first in the list had DDDB and CondDB defined, or not
        # if not, we get the steps of all the previous productions
        if steps[0][4] == steps[0][5] == "fromPreviousStep":
            # if we are here it is because in the current production none of the steps contain DB tags
            self.log.verbose(
                "DB tags are not set: they will be retrieved from the parent production(s)", f"(prod: {prodID})"
            )
            numberOfSteps = len(steps)
            # Now finding the previous productions
            res = self._getPreviousProductions(prodID)
            if not res["OK"]:
                return res
            ancestorProdIDs = res["Value"]  # an already ordered list
            if not ancestorProdIDs:
                return S_ERROR("No ancestor productions found")

            for ancestorProdID in ancestorProdIDs:
                res = self._getRPC().getSteps(ancestorProdID)
                if not res["OK"]:
                    return res
                stepsInAncestorProd = res["Value"]
                steps = stepsInAncestorProd + steps
            allResolvedSteps = self._resolveProductionSteps(steps)
            return S_OK(allResolvedSteps[-numberOfSteps:])
        return S_OK(self._resolveProductionSteps(steps))

    def _resolveProductionSteps(self, steps):
        """Takes care of resolving the steps of a single production
        (including resolving the DDDB and CondDB tags "fromPreviousStep")

        :param list steps: list of steps (which are tuples)
        :returns: list of resolved steps
        """
        if not steps:
            return []

        # DDDB and CondDB are often registered as "fromPreviousStep", so they should be resolved
        # This will search among the steps in the current production (might not be final)
        # A fair assumption is that dddb and conddb are both either set, or not (so both 'fromPreviousStep').

        productionSteps = [steps[0]]
        for i, nextStep in enumerate(steps[1:]):
            nextStep = list(nextStep)
            if nextStep[4] == "fromPreviousStep":
                nextStep[4] = productionSteps[i][4]
            if nextStep[5] == "fromPreviousStep":
                nextStep[5] = productionSteps[i][5]
            productionSteps.append(tuple(nextStep))
        return productionSteps

    def _getPreviousProductions(self, prodID):
        """Returns an already-ordered list of production(s)
        that were inputs to the provided one

        :param str/int prodID: production (transformation) ID
        :returns: S_OK with list of ancestorProdIDs or S_ERROR
        """
        # Start by getting the RequestID
        res = TransformationClient().getTransformation(prodID, True)
        if not res["OK"]:
            self.log.error("Could not retrieve parameters for production", f"{prodID}: {res['Message']}")
            return res
        parameters = res["Value"]

        if parameters["Status"] in ("Cleaned", "Deleted"):
            self.log.notice("The production is Cleaned/Deleted")
            return S_ERROR("The production is Cleaned/Deleted")

        # Now getting the TransformationIDs for the RequestID
        reqID = parameters.get("RequestID")
        if not reqID:
            self.log.error("No RequestID recorded for production", prodID)
            return S_ERROR("No RequestID recorded for production")

        res = ProductionRequestClient().getProductionList(int(reqID))
        if not res["OK"]:
            self.log.error("Could not retrieve productions list for request", f"{reqID}: {res['Message']}")
            return res
        ancestorProdIDs = res["Value"]

        # Now only returning the production IDs prior to prodID (in good order)
        ancestorProdIDs = ancestorProdIDs[0 : ancestorProdIDs.index(prodID)]
        ancestorProdIDs.reverse()
        return S_OK(ancestorProdIDs)


class BKClientWithRetry:
    """Utility class wrapping BKClient with retries."""

    def __init__(self, bkClient=None, retries=None):
        if not bkClient:
            bkClient = BookkeepingClient()
        self.bk = bkClient
        self.retries = retries if retries else 5
        self.method = None

    def __getattr__(self, x):
        self.method = x
        return self.__executeMethod

    def __executeMethod(self, *args, **kwargs):
        fcn = getattr(self.bk, self.method)
        for _i in range(self.retries):
            res = fcn(*args, **kwargs)
            if res["OK"]:
                break
        return res
